/*******************************************************************************
**
** FILE NAME    : ifx_ppa_api_session.c
** PROJECT      : PPA
** MODULES      : PPA API (Routing/Bridging Acceleration APIs)
**
** DATE         : 4 NOV 2008
** AUTHOR       : Xu Liang
** DESCRIPTION  : PPA Protocol Stack Hook API Session Operation Functions
** COPYRIGHT    :              Copyright (c) 2009
**                          Lantiq Deutschland GmbH
**                   Am Campeon 3; 85579 Neubiberg, Germany
**
**   For licensing information, see the file 'LICENSE' in the root folder of
**   this software module.
**
** HISTORY
** $Date        $Author         $Comment
** 04 NOV 2008  Xu Liang        Initiate Version
*******************************************************************************/



/*
 * ####################################
 *              Head File
 * ####################################
 */

/*
 *  Common Head File
 */
#include <linux/kernel.h>
#include <linux/autoconf.h>
#include <linux/netdevice.h>
#include <linux/vmalloc.h>
#include <linux/types.h>
#include <linux/spinlock.h>
//#include <linux/kernel.h>
//#include <linux/module.h>
//#include <linux/version.h>
//#include <linux/types.h>
//#include <linux/init.h>
//#include <linux/slab.h>
//#if defined(CONFIG_IFX_PPA_API_PROC)
//#include <linux/proc_fs.h>
//#endif
//#include <linux/netdevice.h>
//#include <linux/in.h>
//#include <net/sock.h>
//#include <net/ip_vs.h>
//#include <asm/time.h>
#include <net/ipip.h>

/*
 *  PPA Specific Head File
 */
#include <net/ifx_ppa_api.h>
#include <net/ifx_ppa_ppe_hal.h>
#include "ifx_ppa_api_misc.h"
#include "ifx_ppa_api_netif.h"
#include "ifx_ppa_api_session.h"
#include "ifx_ppa_api_pwm.h"

#include "ifx_ppe_drv_wrapper.h"
#if defined(CONFIG_IFX_PMCU) || defined(CONFIG_IFX_PMCU_MODULE)
#include "ifx_ppa_api_pwm.h"
#endif




/*
 * ####################################
 *              Definition
 * ####################################
 */

#define DEFAULT_AVM_HIT_POLLING_TIME            1      //  10 seconds

/*
 *  hash calculation
 */

#define SESSION_LIST_HASH_BIT_LENGTH            10
#define SESSION_LIST_HASH_TABLE_SIZE            ( 1 << SESSION_LIST_HASH_BIT_LENGTH ) // 1024
#define SESSION_LIST_HASH_MASK                  ( SESSION_LIST_HASH_TABLE_SIZE - 1)  // 0x3FF
#define SESSION_LIST_HASH_VALUE(id)				( id & SESSION_LIST_HASH_MASK )

#define SESSION_LIST_MC_HASH_SHIFT              0
#define SESSION_LIST_MC_HASH_BIT_LENGTH         6
#define SESSION_LIST_MC_HASH_MASK               ((1 << SESSION_LIST_MC_HASH_BIT_LENGTH) - 1)
#define SESSION_LIST_MC_HASH_TABLE_SIZE         (1 << SESSION_LIST_MC_HASH_BIT_LENGTH)
#define SESSION_LIST_MC_HASH_VALUE(x)           (((uint32_t)(x) >> SESSION_LIST_MC_HASH_SHIFT) & SESSION_LIST_MC_HASH_MASK)

#define BRIDGING_SESSION_LIST_HASH_BIT_LENGTH   8
#define BRIDGING_SESSION_LIST_HASH_MASK         ((1 << BRIDGING_SESSION_LIST_HASH_BIT_LENGTH) - 1)
#define BRIDGING_SESSION_LIST_HASH_TABLE_SIZE   (1 << BRIDGING_SESSION_LIST_HASH_BIT_LENGTH)
#define BRIDGING_SESSION_LIST_HASH_VALUE(x)     ( ( ((uint32_t)((uint8_t *)(x))[4] << 8) | ((uint8_t *)(x))[5] ) & BRIDGING_SESSION_LIST_HASH_MASK )


/*------------------------------------------------------------------------------------------*\
 * setup use_avm_dest_port_map-Parameter: kann gesetzt werden in:
 *  - /sys/module/ifx_ppa_api/parameters/dest_port_map u.
 *  - kernel_cmd_line: ifx_ppa_api.dest_port_map
\*------------------------------------------------------------------------------------------*/

#if defined(CONFIG_AR9)
int use_avm_dest_port_map = 0;
#else
int use_avm_dest_port_map = 1;
#endif

module_param(use_avm_dest_port_map, int, S_IRUSR | S_IWUSR);

int disable_lan_sessions = 0;
module_param(disable_lan_sessions, int, S_IRUSR | S_IWUSR);

int disable_wlan_egress_sessions = 0;
module_param(disable_wlan_egress_sessions, int, S_IRUSR | S_IWUSR);

int enable_wlan_dscp = 0;
module_param(enable_wlan_dscp , int, S_IRUSR | S_IWUSR);

int enable_wlan_qid = 0;
module_param(enable_wlan_qid , int, S_IRUSR | S_IWUSR);



static DEFINE_RWLOCK( lan_table_lock );
static DEFINE_RWLOCK( wan_table_lock );

#define RW_LOCK_DECLARE unsigned long flags

#define LAN_TABLE_WRITE_LOCK() write_lock_irqsave(&lan_table_lock, flags)
#define WAN_TABLE_WRITE_LOCK() write_lock_irqsave(&wan_table_lock, flags)

#define LAN_TABLE_WRITE_UNLOCK() write_unlock_irqrestore(&lan_table_lock, flags)
#define WAN_TABLE_WRITE_UNLOCK() write_unlock_irqrestore(&wan_table_lock, flags)

#define LAN_TABLE_READ_LOCK() read_lock_irqsave(&lan_table_lock, flags)
#define WAN_TABLE_READ_LOCK() read_lock_irqsave(&wan_table_lock, flags)

#define LAN_TABLE_READ_UNLOCK() read_unlock_irqrestore(&lan_table_lock, flags)
#define WAN_TABLE_READ_UNLOCK() read_unlock_irqrestore(&wan_table_lock, flags)

static avm_session_handle *avm_sessionh_table_lan = NULL;
unsigned int avm_sessionh_table_lan_size = 0;

static avm_session_handle *avm_sessionh_table_wan = NULL;
unsigned int avm_sessionh_table_wan_size = 0;

#if defined(CONFIG_DEBUG_AVM_PPA_VDEV)
unsigned int *avm_session_hit_count_lan_table = NULL;
unsigned int *avm_session_hit_count_wan_table = NULL;

EXPORT_SYMBOL(avm_session_hit_count_lan_table );
EXPORT_SYMBOL(avm_session_hit_count_wan_table );
EXPORT_SYMBOL(avm_sessionh_table_lan_size );
EXPORT_SYMBOL(avm_sessionh_table_wan_size );
#endif


/*
 * ####################################
 *             Declaration
 * ####################################
 */

/*
 *  implemented in PPA PPE Low Level Driver (Data Path)
 */
extern int in_avm_ata_mode(void);

//  routing session list item operation
static INLINE void ppa_init_session_list_item(struct session_list_item *);
/*static INLINE */struct session_list_item *ppa_alloc_session_list_item(void);
/*static INLINE*/ void ppa_free_session_list_item(struct session_list_item *);
/*static INLINE*/ void ppa_insert_session_item(struct session_list_item *);
static INLINE void ppa_remove_session_item(struct session_list_item *);
static void ppa_free_session_list(void);

//  multicast routing group list item operation
void ppa_init_mc_group_list_item(struct mc_group_list_item *);
static INLINE struct mc_group_list_item *ppa_alloc_mc_group_list_item(void);
static INLINE void ppa_free_mc_group_list_item(struct mc_group_list_item *);
static INLINE void ppa_insert_mc_group_item(struct mc_group_list_item *);
static INLINE void ppa_remove_mc_group_item(struct mc_group_list_item *);
static void ppa_free_mc_group_list(void);

static inline void avm_sessionh_table_set(unsigned int rout_table_entry, avm_session_handle sessionh, unsigned int is_lan, unsigned int is_mc );

//  routing session timeout help function
#if !defined(CONFIG_IFX_PPA_AVM_USAGE)
static INLINE uint32_t ppa_get_default_session_timeout(void);
#endif
static void ppa_check_hit_stat(unsigned long);

//  bridging session list item operation
static INLINE void ppa_bridging_init_session_list_item(struct bridging_session_list_item *);
static INLINE struct bridging_session_list_item *ppa_bridging_alloc_session_list_item(void);
static INLINE void ppa_bridging_free_session_list_item(struct bridging_session_list_item *);
static INLINE void ppa_bridging_insert_session_item(struct bridging_session_list_item *);
static INLINE void ppa_bridging_remove_session_item(struct bridging_session_list_item *);
static void ppa_free_bridging_session_list(void);

//  bridging session timeout help function
static INLINE uint32_t ppa_bridging_get_default_session_timeout(void);
static void ppa_bridging_check_hit_stat(unsigned long);

//  help function for special function
static INLINE void ppa_remove_routing_sessions_on_netif(PPA_IFNAME *, uint32_t);
static INLINE void ppa_remove_mc_groups_on_netif(PPA_IFNAME *);
static INLINE void ppa_remove_bridging_sessions_on_netif(PPA_IFNAME *);



/*
 * ####################################
 *           Global Variable
 * ####################################
 */

/*
 *  routing session table
 */
static PPA_LOCK                     g_session_list_lock;
static struct session_list_item    *g_session_list_hash_table[SESSION_LIST_HASH_TABLE_SIZE] = {0};
static uint32_t                     g_session_list_length = 0;
static PPA_MEM_CACHE               *g_session_item_cache = NULL;
static PPA_TIMER                    g_hit_stat_timer;
static uint32_t                     g_hit_polling_time = DEFAULT_AVM_HIT_POLLING_TIME;

/*
 *  multicast routing session table
 */
static PPA_LOCK                     g_mc_group_list_lock;
static struct mc_group_list_item   *g_mc_group_list_hash_table[SESSION_LIST_MC_HASH_TABLE_SIZE] = {0};
static uint32_t                     g_mc_group_list_length = 0;
static PPA_MEM_CACHE               *g_mc_group_item_cache = NULL;

/*
 *  bridging session table
 */
static PPA_LOCK                             g_bridging_session_list_lock;
static struct bridging_session_list_item   *g_bridging_session_list_hash_table[BRIDGING_SESSION_LIST_HASH_TABLE_SIZE] = {0};
static PPA_MEM_CACHE                       *g_bridging_session_item_cache = NULL;
static PPA_TIMER                            g_bridging_hit_stat_timer;
static uint32_t                             g_bridging_hit_polling_time = DEFAULT_BRIDGING_HIT_POLLING_TIME;

int global_session_count = 0;
int global_fail_count = 0;


/*
 * ####################################
 *            Local Function
 * ####################################
 */

/*
 *  routing session list item operation
 */

static INLINE void ppa_init_session_list_item(struct session_list_item *p_item)
{
    ppa_memset(p_item, 0, sizeof(*p_item));
    p_item->mtu             = DEFAULT_MTU;
    p_item->routing_entry   = ~0;
    p_item->pppoe_entry     = ~0;
    p_item->mtu_entry       = ~0;
    p_item->src_mac_entry   = ~0;
    p_item->out_vlan_entry  = ~0;
    p_item->tunnel_idx      = ~0;
}

/*static INLINE */ struct session_list_item *ppa_alloc_session_list_item(void)
{
    struct session_list_item *p_item;

    p_item = ppa_mem_cache_alloc(g_session_item_cache);
    if ( p_item )
        ppa_init_session_list_item(p_item);

    return p_item;
}

/*static INLINE */void ppa_free_session_list_item(struct session_list_item *p_item)
{
    ppa_hw_del_session(p_item);
    ppa_mem_cache_free(p_item, g_session_item_cache);
}

/*static INLINE*/void ppa_insert_session_item(struct session_list_item *p_item)
{
    uint32_t idx;

    idx = SESSION_LIST_HASH_VALUE(p_item->session_id);
    ppa_lock_get(&g_session_list_lock);
    p_item->next = g_session_list_hash_table[idx];
    g_session_list_hash_table[idx] = p_item;
    g_session_list_length++;
    ppa_lock_release(&g_session_list_lock);
}

static INLINE void ppa_remove_session_item(struct session_list_item *p_item)
{
    uint32_t idx;
    struct session_list_item *p_prev, *p_cur;

    idx = SESSION_LIST_HASH_VALUE(p_item->session_id);
    p_prev = NULL;
    ppa_lock_get(&g_session_list_lock);
    p_cur = g_session_list_hash_table[idx];
    while ( p_cur && p_cur != p_item )
    {
        p_prev = p_cur;
        p_cur = p_cur->next;
    }
    if ( p_cur )
    {
        if ( !p_prev )
            g_session_list_hash_table[idx] = p_cur->next;
        else
            p_prev->next = p_cur->next;

        p_cur->next = NULL;

        g_session_list_length--;
    }
    ppa_lock_release(&g_session_list_lock);
}

static void ppa_free_session_list(void)
{
    struct session_list_item *p_item, *p_next;
    int i;

    ppa_lock_get(&g_session_list_lock);
    for ( i = 0; i < NUM_ENTITY(g_session_list_hash_table); i++ )
    {
        for ( p_item = g_session_list_hash_table[i]; p_item; p_item = p_next )
        {
            p_next = p_item->next;

            ppa_free_session_list_item(p_item);
        }
        g_session_list_hash_table[i] = NULL;
    }
    ppa_lock_release(&g_session_list_lock);
}

/*
 *  multicast routing group list item operation
 */

void ppa_init_mc_group_list_item(struct mc_group_list_item *p_item)
{
    ppa_memset(p_item, 0, sizeof(*p_item));
    p_item->mc_entry        = ~0;
    p_item->src_mac_entry   = ~0;
    p_item->out_vlan_entry  = ~0;
}

static INLINE struct mc_group_list_item *ppa_alloc_mc_group_list_item(void)
{
    struct mc_group_list_item *p_item;

    p_item = ppa_mem_cache_alloc(g_mc_group_item_cache);
    if ( p_item )
        ppa_init_mc_group_list_item(p_item);

    return p_item;
}

static INLINE void ppa_free_mc_group_list_item(struct mc_group_list_item *p_item)
{
    ppa_hw_del_mc_group(p_item);
    ppa_mem_cache_free(p_item, g_mc_group_item_cache);
}

static INLINE void ppa_insert_mc_group_item(struct mc_group_list_item *p_item)
{
    uint32_t idx;

    idx = SESSION_LIST_MC_HASH_VALUE(p_item->ip_mc_group.ip.ip);
    ppa_lock_get(&g_mc_group_list_lock);
    p_item->next = g_mc_group_list_hash_table[idx];
    g_mc_group_list_hash_table[idx] = p_item;
    g_mc_group_list_length++;
    ppa_lock_release(&g_mc_group_list_lock);
}

static INLINE void ppa_remove_mc_group_item(struct mc_group_list_item *p_item)
{
    uint32_t idx;
    struct mc_group_list_item *p_prev, *p_cur;

    idx = SESSION_LIST_MC_HASH_VALUE(p_item->ip_mc_group.ip.ip);
    p_prev = NULL;
    ppa_lock_get(&g_mc_group_list_lock);
    p_cur = g_mc_group_list_hash_table[idx];
    while ( p_cur && p_cur != p_item )
    {
        p_prev = p_cur;
        p_cur = p_cur->next;
    }
    if ( p_cur )
    {
        if ( !p_prev )
            g_mc_group_list_hash_table[idx] = p_cur->next;
        else
            p_prev->next = p_cur->next;

        p_cur->next = NULL;

        g_mc_group_list_length--;
    }
    ppa_lock_release(&g_mc_group_list_lock);
}

static void ppa_free_mc_group_list(void)
{
    struct mc_group_list_item *p_mc_item, *p_mc_next;
    int i;

    ppa_lock_get(&g_mc_group_list_lock);
    for ( i = 0; i < NUM_ENTITY(g_mc_group_list_hash_table); i++ )
    {
        for ( p_mc_item = g_mc_group_list_hash_table[i]; p_mc_item; p_mc_item = p_mc_next )
        {
            p_mc_next = p_mc_item->next;

            ppa_free_mc_group_list_item(p_mc_item);
        }
        g_mc_group_list_hash_table[i] = NULL;
    }
    ppa_lock_release(&g_mc_group_list_lock);
}

#if !defined(CONFIG_IFX_PPA_AVM_USAGE)
/*
 *  routing session timeout help function
 */

static INLINE uint32_t ppa_get_default_session_timeout(void)
{
    return DEFAULT_TIMEOUT_IN_SEC;
}
#endif
/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
#if 0
extern void dump_wan_routing_entry(uint32_t entry);

void dump_sessions_with_src_ip_local(unsigned int src_v4_ip){
    unsigned int i;
    struct session_list_item *p_item;
    ppa_lock_get( &g_session_list_lock );
    for ( i = 0; i < NUM_ENTITY(g_session_list_hash_table); i++ ) {
        for ( p_item = g_session_list_hash_table[i]; p_item; p_item = p_item->next ) {
            if( p_item->src_ip.ip == src_v4_ip ) {
               	printk(KERN_ERR "[%s] SRC_IP %#x in routing_entry %d (%s) \n", __FUNCTION__, src_v4_ip, ( p_item->routing_entry & 0x3FFFFFFF ) , ( p_item->flags & SESSION_LAN_ENTRY )?"LAN":"WAN");
               	if (!((1 << 31) & p_item->routing_entry))
               	    dump_wan_routing_entry(p_item->routing_entry & 0x3FFFFFFF );
            }
        }
    }

    ppa_lock_release( &g_session_list_lock );

}
#endif
/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/


avm_session_handle avm_session_handle_slow_lookup(unsigned int routing_entry, unsigned int is_lan, unsigned int is_mc) {
    struct session_list_item *p_item;
    uint32_t i;

    if( is_mc ){
        ppa_debug(DBG_ENABLE_MASK_AVM_PPA, "[%s] multicast sessions are not supported\n");
        return INVALID_SESSION;
    }

    ppa_lock_get( &g_session_list_lock );
    for ( i = 0; i < NUM_ENTITY(g_session_list_hash_table); i++ ) {
        for ( p_item = g_session_list_hash_table[i]; p_item; p_item = p_item->next ) {
            if( p_item->routing_entry == ( ( is_lan << 31) | routing_entry ) ){
               	ppa_debug(DBG_ENABLE_MASK_VDEV_SESSION, "[%s] slow_lookup:avm_session_handle_found=%d routing_entry=%d\n", __FUNCTION__,p_item->session_id, routing_entry);
                ppa_lock_release( &g_session_list_lock );
                return p_item->session_id;
            }
        }
    }

    ppa_lock_release( &g_session_list_lock );
   	ppa_debug(DBG_ENABLE_MASK_AVM_PPA, "[%s] no avm_session_handle_found routing_entry=%d (%s)\n", __FUNCTION__, routing_entry, is_lan?"LAN":"WAN");
   	ppa_debug(DBG_ENABLE_MASK_VDEV_SESSION, "%lu: avm_session_handle not_found routing_entry=%d (%s)\n", jiffies, routing_entry, is_lan?"LAN":"WAN");

#if 0
   	if ( !is_lan && (g_ppa_dbg_enable & DBG_ENABLE_MASK_VDEV_SESSION ) )
        dump_wan_routing_entry(routing_entry);
#endif

    return INVALID_SESSION;

}

/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/

static void ppa_check_hit_stat(unsigned long dummy) {
    struct session_list_item *p_item;
    uint32_t i;
    PPE_ROUTING_INFO route={0};

#ifdef CONFIG_AVM_PA
    if(!in_atomic() && !in_interrupt() && !mutex_trylock( &g_session_list_lock )){
        // still WAITING for AVM_PA to consume session_reports
        ppa_debug(DBG_ENABLE_MASK_AVM_PPA, "[%s] still WAITING for AVM_PA to consume session_reports, in_atomic=%d, in_interrupt=%d\n", __FUNCTION__,  in_atomic(), in_interrupt() );
        ppa_timer_add(&g_hit_stat_timer, 1);
        return;
    }
#endif

    // printk( KERN_ERR "[%s] got LOCK!\n", __func__);
    for ( i = 0; i < NUM_ENTITY(g_session_list_hash_table); i++ )
    {
        for ( p_item = g_session_list_hash_table[i]; p_item; p_item = p_item->next )
        {
            uint64_t bytes_since_last_report = 0;
            route.entry = p_item->routing_entry;
            ifx_ppa_drv_test_and_clear_hit_stat( &route, 0);
            if ( route.f_hit )   {
                ppa_debug(DBG_ENABLE_MASK_VDEV_SESSION, "%lu: hit routing_entry=%d (%s)\n", jiffies, p_item->routing_entry & 0x7FFFFFFF,( p_item->flags & SESSION_LAN_ENTRY )?"LAN":"WAN");
                ppa_debug(DBG_ENABLE_MASK_AVM_PPA, "[%s] session %d was hit\n", __FUNCTION__, p_item->session_id );
                p_item->last_hit_time = ppa_get_time_in_sec();                
                ifx_ppa_drv_get_routing_entry_bytes(&route, 0);

                if( route.bytes >= p_item->last_bytes) 
                    bytes_since_last_report = route.bytes - p_item->last_bytes;
                else
                    bytes_since_last_report = route.bytes + 0xFFFFFFFF - p_item->last_bytes;
                p_item->last_bytes = route.bytes;
                p_item->acc_bytes += bytes_since_last_report;
#ifdef CONFIG_AVM_PA
                if ( p_item->session_id ){
                    // send report to avm_pa (will acquire AVM_PA_LOCK (irqs disabled))
                    avm_pa_hardware_session_report( p_item->session_id, 1, bytes_since_last_report );

                }
#endif
            }
        }
    }
    ppa_lock_release( &g_session_list_lock );

    ppa_timer_add(&g_hit_stat_timer, g_hit_polling_time);
    //printk("jiffy=%lu, HZ=%lu, poll_timer=%lu\n", jiffies, HZ, ppa_api_get_session_poll_timer());
}
/*
 *  bridging session list item operation
 */

static INLINE void ppa_bridging_init_session_list_item(struct bridging_session_list_item *p_item)
{
    ppa_memset(p_item, 0, sizeof(*p_item));
    p_item->bridging_entry = ~0;
}

static INLINE struct bridging_session_list_item *ppa_bridging_alloc_session_list_item(void)
{
    struct bridging_session_list_item *p_item;

    p_item = ppa_mem_cache_alloc(g_bridging_session_item_cache);
    if ( p_item )
        ppa_bridging_init_session_list_item(p_item);

    return p_item;
}

static INLINE void ppa_bridging_free_session_list_item(struct bridging_session_list_item *p_item)
{
    ppa_bridging_hw_del_session(p_item);
    ppa_mem_cache_free(p_item, g_bridging_session_item_cache);
}

static INLINE void ppa_bridging_insert_session_item(struct bridging_session_list_item *p_item)
{
    uint32_t idx;

    idx = BRIDGING_SESSION_LIST_HASH_VALUE(p_item->mac);
    ppa_lock_get(&g_bridging_session_list_lock);
    p_item->next = g_bridging_session_list_hash_table[idx];
    g_bridging_session_list_hash_table[idx] = p_item;
    ppa_lock_release(&g_bridging_session_list_lock);
}

static INLINE void ppa_bridging_remove_session_item(struct bridging_session_list_item *p_item)
{
    uint32_t idx;
    struct bridging_session_list_item *p_prev, *p_cur;

    idx = BRIDGING_SESSION_LIST_HASH_VALUE(p_item->mac);
    p_prev = NULL;
    ppa_lock_get(&g_bridging_session_list_lock);
    p_cur = g_bridging_session_list_hash_table[idx];
    while ( p_cur && p_cur != p_item )
    {
        p_prev = p_cur;
        p_cur = p_cur->next;
    }
    if ( p_cur )
    {
        if ( !p_prev )
            g_bridging_session_list_hash_table[idx] = p_cur->next;
        else
            p_prev->next = p_cur->next;

        p_cur->next = NULL;
    }
    ppa_lock_release(&g_bridging_session_list_lock);
}

static void ppa_free_bridging_session_list(void)
{
    struct bridging_session_list_item *p_brg_item, *p_brg_next;
    int i;

    ppa_lock_get(&g_bridging_session_list_lock);
    for ( i = 0; i < NUM_ENTITY(g_bridging_session_list_hash_table); i++ )
    {
        for ( p_brg_item = g_bridging_session_list_hash_table[i]; p_brg_item; p_brg_item = p_brg_next )
        {
            p_brg_next = p_brg_item->next;

            ppa_bridging_free_session_list_item(p_brg_item);
        }
        g_bridging_session_list_hash_table[i] = NULL;
    }
    ppa_lock_release(&g_bridging_session_list_lock);
}

/*
 *  bridging session timeout help function
 */

static INLINE uint32_t ppa_bridging_get_default_session_timeout(void)
{
    return DEFAULT_BRIDGING_TIMEOUT_IN_SEC;
}

static void ppa_bridging_check_hit_stat(unsigned long dummy)
{
    struct bridging_session_list_item *p_item;
    uint32_t i;
    PPE_BR_MAC_INFO br_mac={0};

    ppa_lock_get(&g_bridging_session_list_lock);
    for ( i = 0; i < NUM_ENTITY(g_bridging_session_list_hash_table); i++ )
    {
        for ( p_item = g_bridging_session_list_hash_table[i]; p_item; p_item = p_item->next )
        {
            br_mac.p_entry = p_item->bridging_entry;
            if ( !(p_item->flags & SESSION_STATIC) )
            {
                ifx_ppa_drv_test_and_clear_bridging_hit_stat( &br_mac, 0);
                if( br_mac.f_hit ) 
                    p_item->last_hit_time = ppa_get_time_in_sec();
            }
        }
    }
    ppa_lock_release(&g_bridging_session_list_lock);

    ppa_timer_add(&g_bridging_hit_stat_timer, g_bridging_hit_polling_time);
}

/*
 *  help function for special function
 */

static INLINE void ppa_remove_routing_sessions_on_netif(PPA_IFNAME *ifname, uint32_t lan_wan_flag)
{
    uint32_t idx;
    struct session_list_item *p_prev, *p_cur;
    struct session_list_item *p_list = NULL, *p_to_del;

    if ( lan_wan_flag == 3 )
    {
        ppa_lock_get(&g_session_list_lock);
        for ( idx = 0; idx < NUM_ENTITY(g_session_list_hash_table); idx++ )
        {
            p_prev = NULL;
            p_cur = g_session_list_hash_table[idx];
            while ( p_cur )
            {
                if ( ppa_is_netif_name(p_cur->rx_if, ifname) || ppa_is_netif_name(p_cur->tx_if, ifname) )
                {
                    p_to_del = p_cur;
                    p_cur = p_cur->next;
                    if ( !p_prev )
                        g_session_list_hash_table[idx] = p_cur;
                    else
                        p_prev->next = p_cur;
                    g_session_list_length--;
                    p_to_del->next = p_list;
                    p_list = p_to_del;
                }
                else
                {
                    p_prev = p_cur;
                    p_cur = p_cur->next;
                }
            }
            ppa_lock_release(&g_session_list_lock);
        }
    }
    else
    {
        uint32_t rx_netif_flag, tx_netif_flag;

        rx_netif_flag = lan_wan_flag == 1 ? SESSION_LAN_ENTRY : SESSION_WAN_ENTRY;
        tx_netif_flag = lan_wan_flag == 1 ? SESSION_WAN_ENTRY : SESSION_LAN_ENTRY;

        ppa_lock_get(&g_session_list_lock);
        for ( idx = 0; idx < NUM_ENTITY(g_session_list_hash_table); idx++ )
        {
            p_prev = NULL;
            p_cur = g_session_list_hash_table[idx];
            while ( p_cur )
            {
                if ( (ppa_is_netif_name(p_cur->rx_if, ifname) && (p_cur->flags & rx_netif_flag))
                    || (ppa_is_netif_name(p_cur->tx_if, ifname) && (p_cur->flags & tx_netif_flag)) )
                {
                    p_to_del = p_cur;
                    p_cur = p_cur->next;
                    if ( !p_prev )
                        g_session_list_hash_table[idx] = p_cur;
                    else
                        p_prev->next = p_cur;
                    g_session_list_length--;
                    p_to_del->next = p_list;
                    p_list = p_to_del;
                }
                else
                {
                    p_prev = p_cur;
                    p_cur = p_cur->next;
                }
            }
            ppa_lock_release(&g_session_list_lock);
        }
    }

    for ( p_to_del = p_list; p_to_del; p_to_del = p_list )
    {
        p_list = p_list->next;
        ppa_free_session_list_item(p_to_del);
    }
}

static INLINE void ppa_remove_mc_groups_on_netif(PPA_IFNAME *ifname)
{
    uint32_t idx;
    struct mc_group_list_item *p_prev, *p_cur;
    struct mc_group_list_item *p_list = NULL, *p_to_del;
    int i, j;

    ppa_lock_get(&g_mc_group_list_lock);
    for ( idx = 0; idx < NUM_ENTITY(g_mc_group_list_hash_table); idx++ )
    {
        p_prev = NULL;
        p_cur = g_mc_group_list_hash_table[idx];
        while ( p_cur )
        {
            for ( i = 0; i < p_cur->num_ifs; i++ )
                if ( ppa_is_netif_name(p_cur->netif[i], ifname) )
                    p_cur->netif[i] = NULL;
            for ( i = j = 0; i < p_cur->num_ifs; i++ )
                if ( p_cur->netif[i] != NULL )
                {
                    if ( i != j )
                    {
                        p_cur->netif[j] = p_cur->netif[i];
                        p_cur->ttl[j] = p_cur->ttl[i];
                    }
                    j++;
                }

            if ( j != 0 )
            {
                p_cur->num_ifs = j;
                p_prev = p_cur;
                p_cur = p_cur->next;
            }
            else
            {
                p_to_del = p_cur;
                p_cur = p_cur->next;
                if ( !p_prev )
                    g_mc_group_list_hash_table[idx] = p_cur;
                else
                    p_prev->next = p_cur;
                g_mc_group_list_length--;
                p_to_del->next = p_list;
                p_list = p_to_del;
            }
        }
    }
    ppa_lock_release(&g_mc_group_list_lock);

    for ( p_to_del = p_list; p_to_del; p_to_del = p_list )
    {
        p_list = p_list->next;
        ppa_free_mc_group_list_item(p_to_del);
    }
}

static INLINE void ppa_remove_bridging_sessions_on_netif(PPA_IFNAME *ifname)
{
    uint32_t idx;
    struct bridging_session_list_item *p_prev, *p_cur;
    struct bridging_session_list_item *p_list = NULL, *p_to_del;

    ppa_lock_get(&g_bridging_session_list_lock);
    for ( idx = 0; idx < NUM_ENTITY(g_bridging_session_list_hash_table); idx++ )
    {
        p_prev = NULL;
        p_cur = g_bridging_session_list_hash_table[idx];
        while ( p_cur )
        {
            if ( ppa_is_netif_name(p_cur->netif, ifname) )
            {
                p_to_del = p_cur;
                p_cur = p_cur->next;
                if ( !p_prev )
                    g_bridging_session_list_hash_table[idx] = p_cur;
                else
                    p_prev->next = p_cur;
                p_to_del->next = p_list;
                p_list = p_to_del;
            }
            else
            {
                p_prev = p_cur;
                p_cur = p_cur->next;
            }
        }
    }
    ppa_lock_release(&g_bridging_session_list_lock);

    for ( p_to_del = p_list; p_to_del; p_to_del = p_list )
    {
        p_list = p_list->next;
        ppa_bridging_free_session_list_item(p_to_del);
    }
}

#if defined(CONFIG_AVM_IFX_PA_ADD_SESSION_PROC)
/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/

static unsigned int parse_flags(const char *flagsstr) {
	unsigned flags = 0;
	if (strstr(flagsstr, "SESSION_INTERNAL_FLAG_BASE"))
		flags |= SESSION_INTERNAL_FLAG_BASE;
	if (strstr(flagsstr, "SESSION_IS_REPLY"))
		flags |= SESSION_IS_REPLY;
	if (strstr(flagsstr, "SESSION_BRIDGING_VCI_CHECK"))
		flags |= SESSION_BRIDGING_VCI_CHECK;
	if (strstr(flagsstr, "SESSION_IS_TCP"))
		flags |= SESSION_IS_TCP;
	if (strstr(flagsstr, "SESSION_ADDED_IN_HW"))
		flags |= SESSION_ADDED_IN_HW;
	if (strstr(flagsstr, "SESSION_CAN_NOT_ACCEL"))
		flags |= SESSION_CAN_NOT_ACCEL;
	if (strstr(flagsstr, "SESSION_STATIC"))
		flags |= SESSION_STATIC;
	if (strstr(flagsstr, "SESSION_DROP"))
		flags |= SESSION_DROP;
	if (strstr(flagsstr, "SESSION_VALID_NAT_IP"))
		flags |= SESSION_VALID_NAT_IP;
	if (strstr(flagsstr, "SESSION_VALID_NAT_PORT"))
		flags |= SESSION_VALID_NAT_PORT;
	if (strstr(flagsstr, "SESSION_VALID_NAT_SNAT"))
		flags |= SESSION_VALID_NAT_SNAT;
	if (strstr(flagsstr, "SESSION_VALID_VLAN_INS"))
		flags |= SESSION_VALID_VLAN_INS;
	if (strstr(flagsstr, "SESSION_VALID_VLAN_RM"))
		flags |= SESSION_VALID_VLAN_RM;
	if (strstr(flagsstr, "SESSION_VALID_OUT_VLAN_INS"))
		flags |= SESSION_VALID_OUT_VLAN_INS;
	if (strstr(flagsstr, "SESSION_VALID_OUT_VLAN_RM"))
		flags |= SESSION_VALID_OUT_VLAN_RM;
	if (strstr(flagsstr, "SESSION_VALID_PPPOE"))
		flags |= SESSION_VALID_PPPOE;
	if (strstr(flagsstr, "SESSION_VALID_NEW_SRC_MAC"))
		flags |= SESSION_VALID_NEW_SRC_MAC;
	if (strstr(flagsstr, "SESSION_VALID_SRC_MAC"))
		flags |= SESSION_VALID_SRC_MAC;
	if (strstr(flagsstr, "SESSION_VALID_MTU"))
		flags |= SESSION_VALID_MTU;
	if (strstr(flagsstr, "SESSION_VALID_NEW_DSCP"))
		flags |= SESSION_VALID_NEW_DSCP;
	if (strstr(flagsstr, "SESSION_VALID_DSLWAN_QID"))
		flags |= SESSION_VALID_DSLWAN_QID;
	if (strstr(flagsstr, "SESSION_TX_ITF_IPOA"))
		flags |= SESSION_TX_ITF_IPOA;
	if (strstr(flagsstr, "SESSION_TX_ITF_PPPOA"))
		flags |= SESSION_TX_ITF_PPPOA;
	if (strstr(flagsstr, "SESSION_TX_ITF_IPOA_PPPOA_MASK"))
		flags |= SESSION_TX_ITF_IPOA_PPPOA_MASK;
	if (strstr(flagsstr, "SESSION_SRC_MAC_DROP_EN"))
		flags |= SESSION_SRC_MAC_DROP_EN;
	if (strstr(flagsstr, "SESSION_LAN_ENTRY"))
		flags |= SESSION_LAN_ENTRY;
	if (strstr(flagsstr, "SESSION_WAN_ENTRY"))
		flags |= SESSION_WAN_ENTRY;
	if (strstr(flagsstr, "SESSION_IS_IPV6"))
		flags |= SESSION_IS_IPV6;
	return flags;
}

/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
static int parse_ip_addr(const char* ipstr) {
	int i;
	int q[] = { -1, -1, -1, -1 };
	int ip;
	sscanf(ipstr, "%d.%d.%d.%d", &q[0], &q[1], &q[2], &q[3]);
	for (i = 0; i < 4; i++) {
		if ((q[i] < 0) || (q[i] > 255)) {
			ppa_debug(DBG_ENABLE_MASK_AVM_PPA, "[%s] invalid ip\n", __FUNCTION__);
			return 0;
		}
	}
	ip = ((q[0] << 24) | (q[1] << 16) | (q[2] << 8) | (q[3] << 0));
	ppa_debug(DBG_ENABLE_MASK_AVM_PPA, "[%s] %#x\n", __FUNCTION__, ip);
	return ip;
}
/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/

static int isdigit(char c) {
	return ((c >= '0') && (c <= '9'));
}

/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/

static int isxdigit(char c) {
	return isdigit(c) || ((c>='a')&&(c<='f')) || ((c>='A')&&(c<='F'));
}
/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/

static int islower(char c) {
    return((c>='a')&&(c<='z'));
}

/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/

static int toupper(char c) {
	if (islower(c))
		return c - ('a' - 'A');
	else
		return c;
}

/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/

static void parse_mac_addr(const char *macstr, uint8_t *mac_addr) {
	int i, j;
	unsigned char result, value;
	const char *iter = macstr;

	for (i = 0; i < 6; i++) {
		result = 0;

		if (i != 5 && *(iter + 2) != ':') {
			ppa_debug(DBG_ENABLE_MASK_AVM_PPA, "[%s] invalid mac i=%d\n", __FUNCTION__, i);
			return;
		}

		for (j = 0; j < 2; j++) {
			if (isxdigit(*iter) && ((value = isdigit(*iter) ? *iter - '0' : toupper(*iter) - 'A' + 10) < 16)) {
				result = result * 16 + value;
				iter++;
			} else {
				ppa_debug(DBG_ENABLE_MASK_AVM_PPA, "[%s] (1) invalid mac i=%d, j=%d, iter='%c'\n", __FUNCTION__, i,j , *iter);
				return;
			}
		}

		iter++;
		mac_addr[i] = result;
	}
	ppa_debug(DBG_ENABLE_MASK_AVM_PPA, "[%s] mac_str=%s mac=%02x %02x %02x %02x %02x %02x\n", __FUNCTION__,
			macstr, mac_addr[0], mac_addr[1], mac_addr[2], mac_addr[3],
			mac_addr[4], mac_addr[5]);
}

/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
static int parse_digit(const char *digitstr) {
	int res = 0;
	sscanf(digitstr, "%d", &res);
	return res;
}
/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
static char *parse_string(const char *string) {
	const char *c;
	int size = 0;
	for (c = string; *c; c++) {
		if ((*c == '\n') || (*c == '\t'))
			break;
	}
	size = c - string;

	if (size > 0) {
		char *res = vmalloc(size + 1);
		if (!res)
			return 0;
		strncpy(res, string, size);
		res[size] = '\0';
		return res;
	}
	return 0;
}

/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
enum parse_res{SESSION_BEGIN,SESSION_END,SESSION_VAL,SESSION_COMMENT,SESSION_UNKNOWN};

int parse_line(const char* line, struct session_list_item *p_item, int *session_id) {
	char *pos;
	*session_id = 0;
	ppa_debug(DBG_ENABLE_MASK_AVM_PPA, "[%s] %s \n", __FUNCTION__, line);

	for (pos = line; pos!=0; pos++ ){
		if (*pos == '#'){
			ppa_debug(DBG_ENABLE_MASK_AVM_PPA, "[%s] matched comment '%s' \n", __FUNCTION__, line);
			return SESSION_COMMENT;
		}
		if (*pos != '\t' && *pos != ' ')
			break;
	}

	if ((pos = strstr(line, "src_ip"))) {
		p_item->src_ip.ip = parse_ip_addr(pos + 10);
		return SESSION_VAL;
	}

	if ((pos = strstr(line, "src_port"))) {
		p_item->src_port = parse_digit(pos + 12);
		return SESSION_VAL;
	}

	if ((pos = strstr(line, "new_src_mac"))) {
		ppa_debug(DBG_ENABLE_MASK_AVM_PPA, "[%s]new_src_mac at %p ", __FUNCTION__, p_item->new_src_mac);
		parse_mac_addr(pos + 15, p_item->new_src_mac);
		return SESSION_VAL;
	}

	if ((pos = strstr(line, "src_mac"))) {
		parse_mac_addr(pos + 11, p_item->src_mac);
		return SESSION_VAL;
	}

	if ((pos = strstr(line, "dst_ip"))) {
		p_item->dst_ip.ip = parse_ip_addr(pos + 10);
		return SESSION_VAL;
	}

	if ((pos = strstr(line, "dst_port"))) {
		p_item->dst_port = parse_digit(pos + 12);
		return SESSION_VAL;
	}

	if ((pos = strstr(line, "dst_mac"))) {
		parse_mac_addr(pos + 11, p_item->dst_mac);
		return SESSION_VAL;
	}

	if ((pos = strstr(line, "nat_ip"))) {
		p_item->nat_ip.ip = parse_ip_addr(pos + 10);
		return SESSION_VAL;
	}

	if ((pos = strstr(line, "nat_port"))) {
		p_item->nat_port = parse_digit(pos + 12);
		return SESSION_VAL;
	}

	if ((pos = strstr(line, "new_dscp"))) {
		p_item->new_dscp = parse_digit(pos + 12);
		return SESSION_VAL;
	}

	if ((pos = strstr(line, "ip_tos"))) {
		p_item->ip_tos = parse_digit(pos + 10);
		return SESSION_VAL;
	}

	if ((pos = strstr(line, "ip_proto"))) {
		int proto = parse_digit(pos + 12);
		ppa_debug(DBG_ENABLE_MASK_AVM_PPA, "[%s] proto =%d", __FUNCTION__, proto);
		p_item->ip_proto = proto;
		return SESSION_VAL;
	}

	if ((pos = strstr(line, "pppoe_session_id"))) {
		p_item->pppoe_session_id = parse_digit(pos + 20);
		return SESSION_VAL;
	}

	if ((pos = strstr(line, "tx_if_name"))) {
		char *tx_if_str = parse_string(pos + 14);
		ppa_debug(DBG_ENABLE_MASK_AVM_PPA, "[%s] tx_if_str='%s'\n", __FUNCTION__, tx_if_str);
		vfree(tx_if_str);
		return SESSION_VAL;
	}

	if ((pos = strstr(line, "new_vci"))) {
		p_item->new_vci = parse_digit(pos + 11);
		return SESSION_VAL;
	}

	if ((pos = strstr(line, "out_vlan_tag"))) {
		p_item->out_vlan_tag = parse_digit(pos + 16);
		return SESSION_VAL;
	}

	if ((pos = strstr(line, "mtu"))) {
		p_item->mtu = parse_digit(pos + 7);
		return SESSION_VAL;
	}

	if ((pos = strstr(line, "dslwan_qid"))) {
		p_item->dslwan_qid = parse_digit(pos + 14);
		return SESSION_VAL;
	}

	if ((pos = strstr(line, "dest_ifid"))) {
		p_item->dest_ifid = parse_digit(pos + 13);
		return SESSION_VAL;
	}

	if ((pos = strstr(line, "flags"))) {
		p_item->flags = parse_flags(pos + 9);
		return SESSION_VAL;
	}

	if ((pos = strstr(line, "skb_priority"))) {
		p_item->priority = parse_digit(pos + 16);
		return SESSION_VAL;
	}

	if ((pos = strstr(line, "session"))) {
		*session_id = parse_digit(pos + 8);
		return SESSION_BEGIN;
	}
	if ((pos = strstr(line, ");"))) {
		return SESSION_END;
	}
	return SESSION_UNKNOWN;
}

/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/

int proc_write_avm_del_session(struct file *file, const char *buf, unsigned long count, void *data) {
	int session_id;
	char *proc_str_buf;
	struct session_list_item *p_item = NULL;

	proc_str_buf = vmalloc(count + 1);
	if (!proc_str_buf ) {
		return 0;
	}
	proc_str_buf[count]= 0;

	if(copy_from_user(proc_str_buf, buf, count)){
		ppa_debug(DBG_ENABLE_MASK_AVM_PPA, "[%s] copy from user failed\n", __FUNCTION__);
		vfree(proc_str_buf);
		return 0;
	}

	session_id = parse_digit(proc_str_buf);
	vfree(proc_str_buf);

	ppa_debug(DBG_ENABLE_MASK_AVM_PPA, "[%s] lookup session id %d \n", __FUNCTION__, session_id);

	//  i = 0, org dir, i = 1, reply dir
    if ( avm_lookup_session(session_id, &p_item) != IFX_PPA_SESSION_EXISTS ){
    	ppa_debug(DBG_ENABLE_MASK_AVM_PPA, "[%s]session not found!\n", __FUNCTION__);
    	return count;
    }

    dump_list_item(p_item, "ifx_ppa_session_delete");

#if 0
    /*--- das wird in ppa_free_session_list_item erledigt ---*/
   	ppa_debug(DBG_ENABLE_MASK_AVM_PPA, "[%s] del session %p\n", __FUNCTION__, p_item);
    ppa_hw_del_session(p_item); 
#endif 
    ppa_debug(DBG_ENABLE_MASK_AVM_PPA, "[%s] remove session %p\n", __FUNCTION__, p_item);
    ppa_remove_session(p_item);

    ppa_debug(DBG_ENABLE_MASK_AVM_PPA, "[%s] deaktivate ppa pwm\n", __FUNCTION__);
    ifx_ppa_pwm_deactivate_module();

	return count;

}

/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/


int proc_write_avm_add_session(struct file *file, const char *buf, unsigned long count, void *data) {
    static int global_session_ppa_counter = 1;
	char *proc_str_buf = NULL;
	int i, token_start;
	int update_old_session = 0;
	struct session_list_item *p_item = NULL;

	proc_str_buf = vmalloc(count + 1);
	if (!proc_str_buf ) {
		return 0;
	}


	ppa_debug(DBG_ENABLE_MASK_AVM_PPA, "[%s] try to copy %d bytes \n", __FUNCTION__, count);
	if(copy_from_user(proc_str_buf, buf, count)){
		ppa_debug(DBG_ENABLE_MASK_AVM_PPA, "[%s] copy from user failed\n", __FUNCTION__);
		vfree(proc_str_buf);
		return 0;
	}
	proc_str_buf[count]= 0;

	ppa_debug(DBG_ENABLE_MASK_AVM_PPA, "[%s] copied %d bytes\n", __FUNCTION__, count);
	ppa_debug(DBG_ENABLE_MASK_AVM_PPA, "'%s'\n", proc_str_buf);

	for (i = 0, token_start = 0; i < count; i++) {
		if (proc_str_buf[i] == '\n') {
			char *new_token;
			int session_id = 0;

			int token_size = i - token_start;
			new_token = vmalloc(token_size + 1);
			if (!new_token) {
				vfree(proc_str_buf);
				vfree(p_item);
				return 0;
			}
			ppa_debug(DBG_ENABLE_MASK_AVM_PPA, "[%s] new token len %d bytes\n", __FUNCTION__, token_size + 1);
			strncpy(new_token, &proc_str_buf[token_start], token_size);
			new_token[token_size] = '\0';

			switch(parse_line(new_token, p_item, &session_id)){

				case SESSION_BEGIN:
					p_item = NULL;
					ppa_debug(DBG_ENABLE_MASK_AVM_PPA, "[%s] SESSION_BEGIN\n", __FUNCTION__);
					if (session_id){
						avm_lookup_session(session_id, &p_item);
					}
					if(p_item){
						ppa_debug(DBG_ENABLE_MASK_AVM_PPA, "[%s] got old session %d\n", __FUNCTION__, session_id);
						ppa_debug(DBG_ENABLE_MASK_AVM_PPA, "[%s] del old item\n", __FUNCTION__);
						ppa_hw_del_session(p_item);
						ppa_debug(DBG_ENABLE_MASK_AVM_PPA, "[%s] deaktivate ppa\n", __FUNCTION__);
						ifx_ppa_pwm_deactivate_module();
						update_old_session = 1;
					} else {
						p_item = ppa_alloc_session_list_item();
						if ( !p_item ) {
							err("failed in memory allocation");
							vfree(proc_str_buf);
							return IFX_ENOMEM;
						}

						ppa_debug(DBG_ENABLE_MASK_AVM_PPA, "[%s] session item allocated, %p\n", __FUNCTION__, p_item);
					}
					if (session_id)
						p_item->session = (PPA_SESSION*)session_id;
					else {
						global_session_ppa_counter++; //TODO brauche wir die?!
						p_item->session = (PPA_SESSION*)global_session_ppa_counter;
					}
					break;

				case SESSION_END:
					ppa_debug(DBG_ENABLE_MASK_AVM_PPA, "[%s] SESSION_END\n", __FUNCTION__);
					if ( !update_old_session ){
						int ret;
						dump_list_item(p_item, "proc_write_avm_add_session (after setting)");
						ppa_insert_session_item(p_item);
						if ( (ret = ppa_hw_add_session(p_item, 0)) != IFX_SUCCESS ) {
							ppa_debug(DBG_ENABLE_MASK_AVM_PPA, "[%s] ppa_hw_add_session failed: %d\n", __FUNCTION__, ret);
						}
					} else {
						int ret;
						ppa_debug(DBG_ENABLE_MASK_AVM_PPA, "[%s] add new item\n", __FUNCTION__);
						if ( (ret = ppa_hw_add_session(p_item, 0)) != IFX_SUCCESS ) {
							ppa_debug(DBG_ENABLE_MASK_AVM_PPA, "[%s] update failed in ppa_hw_add_session : %d\n", __FUNCTION__, ret);
						}
					}
					ifx_ppa_pwm_activate_module();

					update_old_session = 0;
					p_item = NULL;
					break;
				case SESSION_UNKNOWN:
					ppa_debug(DBG_ENABLE_MASK_AVM_PPA, "[%s] could not parse line\n", __FUNCTION__);
					break;
				case SESSION_VAL:
					ppa_debug(DBG_ENABLE_MASK_AVM_PPA, "[%s] parsed line\n", __FUNCTION__);
					break;
			}
			vfree(new_token);
			token_start = i + 1;
		}
	}
	vfree(proc_str_buf);
	return count;
}
#endif //#if defined(CONFIG_AVM_IFX_PA_ADD_SESSION_PROC)



/*
 * ####################################
 *           Global Function
 * ####################################
 */

/*
 *  routing session operation
 */

#if defined(CONFIG_IFX_PPA_AVM_USAGE)

int32_t avm_lookup_session(uint16_t session_id, struct session_list_item **pp_item){
    int32_t ret;
    struct session_list_item *p_prev, *p_cur;
    struct session_list_item **pp_header;

    ASSERT(pp_item != NULL, "pp_item == NULL");

    ret = IFX_PPA_SESSION_NOT_ADDED;

    pp_header = g_session_list_hash_table + SESSION_LIST_HASH_VALUE(session_id);
    p_prev = NULL;
    ppa_lock_get(&g_session_list_lock);
    p_cur = *pp_header;
    while ( p_cur &&  (session_id != p_cur->session_id))
    {
    	p_prev = p_cur;
    	p_cur = p_cur->next;
    }
    if ( p_cur ) // gefunden? dann kommt der eintrag an die erste stelle
    {
    	if ( p_prev )
    	{
    		p_prev->next = p_cur->next;
    		p_cur->next = *pp_header;
    		*pp_header = p_cur;
    	}

    	ret = IFX_PPA_SESSION_EXISTS;
    }
    ppa_lock_release(&g_session_list_lock);

    *pp_item = p_cur;

    return ret;

}
#endif /*--- #if defined(CONFIG_IFX_PPA_AVM_USAGE) ---*/


void ppa_remove_session(struct session_list_item *p_item)
{
    ppa_remove_session_item(p_item);

    ppa_free_session_list_item(p_item);
}

void dump_list_item(struct session_list_item *p_item, char *comment)
{
#if defined(DEBUG_DUMP_LIST_ITEM) && DEBUG_DUMP_LIST_ITEM
	int8_t strbuf[64];
    if ( !(g_ppa_dbg_enable & DBG_ENABLE_MASK_DUMP_ROUTING_SESSION) )
        return;

    if ( comment )
        printk("dump_list_item - %s\n", comment);
    else
        printk("dump_list_item\n");
    printk("  next             = %08X\n", (uint32_t)p_item->next);
    printk("  session          = %d\n", (uint32_t)p_item->session_id);
    printk("  ip_proto         = %08X\n", p_item->ip_proto);
    printk("  src_ip           = %s\n",   ppa_get_pkt_ip_string(p_item->src_ip, p_item->flags & SESSION_IS_IPV6, strbuf));
    printk("  src_port         = %d\n",   p_item->src_port);
    printk("  src_mac[6]       = %s\n",   ppa_get_pkt_mac_string(p_item->src_mac, strbuf));
    printk("  dst_ip           = %s\n",   ppa_get_pkt_ip_string(p_item->dst_ip, p_item->flags & SESSION_IS_IPV6, strbuf));
    printk("  dst_port         = %d\n",   p_item->dst_port);
    printk("  dst_mac[6]       = %s\n",   ppa_get_pkt_mac_string(p_item->dst_mac, strbuf));
    printk("  nat_ip           = %s\n",   ppa_get_pkt_ip_string(p_item->nat_ip, p_item->flags & SESSION_IS_IPV6, strbuf));
    printk("  nat_port         = %d\n",   p_item->nat_port);
    printk("  new_src_mac[6]   = %s\n",   ppa_get_pkt_mac_string(p_item->new_src_mac, strbuf));
    printk("  rx_if            = %08X\n", (uint32_t)p_item->rx_if);
    printk("  tx_if            = %08X\n", (uint32_t)p_item->tx_if);
    printk("  timeout          = %d\n",   p_item->timeout);
    printk("  last_hit_time    = %d\n",   p_item->last_hit_time);
    printk("  num_adds         = %d\n",   p_item->num_adds);
    printk("  pppoe_session_id = %d\n",   p_item->pppoe_session_id);
    printk("  new_dscp         = %d\n",   p_item->new_dscp);
    printk("  new_vci          = %08X\n",  p_item->new_vci);
    printk("  mtu              = %d\n",   p_item->mtu);
    printk("  flags            = %08X\n", p_item->flags);
    printk("  routing_entry    = %08X\n", p_item->routing_entry);
    printk("  pppoe_entry      = %08X\n", p_item->pppoe_entry);
    printk("  mtu_entry        = %08X\n", p_item->mtu_entry);
    printk("  src_mac_entry    = %08X\n", p_item->src_mac_entry);

#endif
}

int32_t ppa_session_start_iteration(uint32_t *ppos, struct session_list_item **pp_item)
{
    struct session_list_item *p = NULL;
    int idx;
    uint32_t l;

    l = *ppos + 1;

    ppa_lock_get(&g_session_list_lock);

    for ( idx = 0; l && idx < NUM_ENTITY(g_session_list_hash_table); idx++ )
    {
        for ( p = g_session_list_hash_table[idx]; p; p = p->next )
            if ( !--l )
                break;
    }

    if ( l == 0 && p )
    {
        ++*ppos;
        *pp_item = p;
        return IFX_SUCCESS;
    }
    else
    {
        *pp_item = NULL;
        return IFX_FAILURE;
    }
}

int32_t ppa_session_iterate_next(uint32_t *ppos, struct session_list_item **pp_item)
{
    uint32_t idx;

    if ( *pp_item == NULL )
        return IFX_FAILURE;

    if ( (*pp_item)->next != NULL )
    {
        ++*ppos;
        *pp_item = (*pp_item)->next;
        return IFX_SUCCESS;
    }
    else
    {
        for ( idx = SESSION_LIST_HASH_VALUE((*pp_item)->session_id) + 1; idx < NUM_ENTITY(g_session_list_hash_table); idx++ )
            if ( g_session_list_hash_table[idx] != NULL )
            {
                ++*ppos;
                *pp_item = g_session_list_hash_table[idx];
                return IFX_SUCCESS;
            }
        *pp_item = NULL;
        return IFX_FAILURE;
    }
}

void ppa_session_stop_iteration(void)
{
    ppa_lock_release(&g_session_list_lock);
}

/*------------------------------------------------------------------------------------------*\
 * ppa_hw_add_session aus ugw 5.1.1
\*------------------------------------------------------------------------------------------*/


/*
 *  routing session hardware/firmware operation
 */
int32_t ppa_hw_add_session(struct session_list_item *p_item, uint32_t f_test)
{
    PPE_ROUTING_INFO route={0};
    
    int ret = IFX_SUCCESS;

    //  Only add session in H/w when the called from the post-NAT hook
    ppa_memset( &route, 0, sizeof(route) );
    route.src_mac.mac_ix = ~0;
    route.pppoe_info.pppoe_ix = ~0;
    route.out_vlan_info.vlan_entry = ~0;
    route.mtu_info.mtu_ix = ~0;
    
    route.route_type = (p_item->flags & SESSION_VALID_NAT_IP) ? ((p_item->flags & SESSION_VALID_NAT_PORT) ? IFX_PPA_ROUTE_TYPE_NAPT : IFX_PPA_ROUTE_TYPE_NAT) : IFX_PPA_ROUTE_TYPE_IPV4;
    if ( (p_item->flags & SESSION_VALID_NAT_IP) )
    {
        route.new_ip.f_ipv6 = 0;
        memcpy( &route.new_ip.ip, &p_item->nat_ip, sizeof(route.new_ip.ip) ); //since only IPv4 support NAT, translate it to IPv4 format
    }
    if ( (p_item->flags & SESSION_VALID_NAT_PORT) )
        route.new_port = p_item->nat_port;

    ppa_debug(DBG_ENABLE_MASK_AVM_PPA, "[%s] flags=%#x\n", __FUNCTION__, p_item->flags);
    if ( (p_item->flags & (SESSION_VALID_PPPOE | SESSION_LAN_ENTRY)) == (SESSION_VALID_PPPOE | SESSION_LAN_ENTRY) )
    {
    	ppa_debug(DBG_ENABLE_MASK_AVM_PPA, "[%s] try to add pppoe_session_id=%d\n", __FUNCTION__, p_item->pppoe_session_id);
        route.pppoe_info.pppoe_session_id = p_item->pppoe_session_id;
        if ( ifx_ppa_drv_add_pppoe_entry( &route.pppoe_info, 0) == IFX_SUCCESS ){
        	ppa_debug(DBG_ENABLE_MASK_AVM_PPA, "[%s] try to added pppoe_ix=%d\n", __FUNCTION__, route.pppoe_info.pppoe_ix);
            p_item->pppoe_entry = route.pppoe_info.pppoe_ix;
        }
        else {
            dbg("add pppoe_entry error");
            SET_DBG_FLAG(p_item, SESSION_DBG_ADD_PPPOE_ENTRY_FAIL);
            goto SESSION_VALID_PPPOE_ERROR;
        }
    }

    route.mtu_info.mtu = p_item->mtu;
    if ( ifx_ppa_drv_add_mtu_entry(&route.mtu_info, 0) == IFX_SUCCESS )
    {
        p_item->mtu_entry = route.mtu_info.mtu_ix;
        p_item->flags |= SESSION_VALID_MTU;
    }
    else
    {
        SET_DBG_FLAG(p_item, SESSION_DBG_ADD_MTU_ENTRY_FAIL);
        goto MTU_ERROR;
    }

    if((p_item->flags & (SESSION_TUNNEL_6RD | SESSION_LAN_ENTRY)) == (SESSION_TUNNEL_6RD | SESSION_LAN_ENTRY)){
    	if ( p_item->tun_daddr.ip && p_item->tun_saddr.ip ){
    		// AVM-WAY
			dbg( "add 6RD entry to FW (avm-way) saddr %#x, daddr %#x\n", p_item->tun_saddr.ip, p_item->tun_daddr.ip );
			route.tnnl_info.tunnel_type = SESSION_TUNNEL_6RD;
			route.tnnl_info.saddr =  p_item->tun_saddr.ip;
			route.tnnl_info.daddr =  p_item->tun_daddr.ip;
    	} else {
			struct ip_tunnel *tun;
			ASSERT(p_item->tx_if->type == ARPHRD_SIT, "device is not a sit device");
			dbg("add 6RD entry to FW, tx dev: %s", p_item->tx_if->name);
			route.tnnl_info.tunnel_type = SESSION_TUNNEL_6RD;

			tun = (struct ip_tunnel *)netdev_priv(p_item->tx_if);
			route.tnnl_info.saddr =  tun->parms.iph.saddr;
			route.tnnl_info.daddr =  tun->parms.iph.daddr;

    	}

        if(ifx_ppa_drv_add_tunnel_entry(&route.tnnl_info, 0) == IFX_SUCCESS){
            p_item->tunnel_idx = route.tnnl_info.tunnel_idx;
        }else{
            dbg("add tunnel 6rd entry error");
            goto MTU_ERROR;
        }
    }
    if(p_item->flags & SESSION_TUNNEL_6RD){
        route.tnnl_info.tunnel_idx = p_item->tunnel_idx << 1 | 1;
    }

    if (p_item->flags & SESSION_VALID_NEW_SRC_MAC){

       	// this is the regular avm way
        ppa_memcpy(route.src_mac.mac, p_item->new_src_mac, sizeof(route.src_mac.mac) );

    } else if ( !(p_item->flags & SESSION_TX_ITF_IPOA_PPPOA_MASK)) {

        if( !f_test ) {
        	// this is the regular lantiq way
			ppa_debug(DBG_ENABLE_MASK_AVM_PPA, "[%s] warning doing mac lookup!\n", __FUNCTION__ );
            ppa_get_netif_hwaddr(p_item->tx_if, route.src_mac.mac);

        } else {

            //for testing only: used for ioctl to add a fake routing accleration entry in PPE 
			ppa_debug(DBG_ENABLE_MASK_AVM_PPA, "[%s] warning test mode!\n", __FUNCTION__ );
            ppa_memcpy(route.src_mac.mac, p_item->src_mac, sizeof(route.src_mac.mac) );

        }
        p_item->flags |= SESSION_VALID_NEW_SRC_MAC;
    }

    if (p_item->flags & SESSION_VALID_NEW_SRC_MAC){
    	// Die new_src_mac wird redundant abgelegt
        ppa_memcpy( p_item->new_src_mac, route.src_mac.mac,  sizeof(route.src_mac.mac) );
        if ( ifx_ppa_drv_add_mac_entry(&route.src_mac, 0) == IFX_SUCCESS )
        {
            p_item->src_mac_entry = route.src_mac.mac_ix;
			ppa_debug(DBG_ENABLE_MASK_AVM_PPA, "[%s] added new mac entry (new_src_mac) ix=%d\n", __FUNCTION__, p_item->src_mac_entry );
        }
        else
        {
            SET_DBG_FLAG(p_item, SESSION_DBG_ADD_MAC_ENTRY_FAIL);
            goto NEW_SRC_MAC_ERROR;
        }
    }


    if ( (p_item->flags & SESSION_VALID_OUT_VLAN_INS) )
    {
        route.out_vlan_info.vlan_id = p_item->out_vlan_tag;
        if ( ifx_ppa_drv_add_outer_vlan_entry( &route.out_vlan_info, 0 ) == IFX_SUCCESS )
        {            
            p_item->out_vlan_entry = route.out_vlan_info.vlan_entry;
        }
        else
        {
            dbg("add out_vlan_ix error");
            SET_DBG_FLAG(p_item, SESSION_DBG_ADD_OUT_VLAN_ENTRY_FAIL);
            goto OUT_VLAN_ERROR;
        }
    }

    if ( (p_item->flags & SESSION_VALID_NEW_DSCP) )
        route.new_dscp = p_item->new_dscp;

    route.dest_list = 1 << p_item->dest_ifid;

    if (use_avm_dest_port_map && p_item->dest_port_map){
    		route.dest_list |= ((p_item->dest_port_map & 0x3F) << 2);
    }

    dbg("merged dest_list = %#x, dest_port_map=%#x", route.dest_list, p_item->dest_port_map);

    route.f_is_lan = p_item->flags & SESSION_LAN_ENTRY;
    ppa_memcpy(route.new_dst_mac, p_item->dst_mac, PPA_ETH_ALEN);
    route.dst_port = p_item->dst_port;
    route.src_port = p_item->src_port;
    route.f_is_tcp = p_item->flags & SESSION_IS_TCP;
    route.f_new_dscp_enable = p_item->flags & SESSION_VALID_NEW_DSCP;
    route.f_vlan_ins_enable =p_item->flags & SESSION_VALID_VLAN_INS;
    route.new_vci = p_item->new_vci;
    route.f_vlan_rm_enable = p_item->flags & SESSION_VALID_VLAN_RM;
    route.pppoe_mode = p_item->flags & SESSION_VALID_PPPOE;
    route.f_out_vlan_ins_enable =p_item->flags & SESSION_VALID_OUT_VLAN_INS;    
    route.f_out_vlan_rm_enable = p_item->flags & SESSION_VALID_OUT_VLAN_RM;
    route.dslwan_qid = p_item->dslwan_qid;
    
#if defined(CONFIG_IFX_PPA_IPv6_ENABLE)     
    if( p_item->flags & SESSION_IS_IPV6 )
    {
        route.src_ip.f_ipv6 = 1;
        ppa_memcpy( route.src_ip.ip.ip6, p_item->src_ip.ip6, sizeof(route.src_ip.ip.ip6));

        route.dst_ip.f_ipv6 = 1;
        ppa_memcpy( route.dst_ip.ip.ip6, p_item->dst_ip.ip6, sizeof(route.dst_ip.ip.ip6)); 
    }else
#endif
    {
        route.src_ip.f_ipv6 = 0;
        route.src_ip.ip.ip= p_item->src_ip.ip;

        route.dst_ip.f_ipv6 = 0;
        route.dst_ip.ip.ip= p_item->dst_ip.ip; 

        route.new_ip.f_ipv6 = 0;
    }

    /*
     * will be activated after session_hash_table update
     * otherwise fist WLAN-Packets will be dropped, as session_id is unkown by kernel yet
     */
    route.active = 0;


    if ( (ret = ifx_ppa_drv_add_routing_entry( &route, 0) ) == IFX_SUCCESS ) {
        p_item->routing_entry = route.entry;
        p_item->flags |= SESSION_ADDED_IN_HW;
        dbg("add_routing_entry successful");
        global_session_count++;

        avm_sessionh_table_set( p_item->routing_entry, p_item->session_id, (p_item->flags & SESSION_LAN_ENTRY), 0);

        ifx_ppa_drv_activate_routing_entry( &route, 0);
        return IFX_SUCCESS;
    }
    

    //  fail in add_routing_entry
    dbg("fail in add_routing_entry");
    p_item->out_vlan_entry = ~0;
    ifx_ppa_drv_del_outer_vlan_entry( &route.out_vlan_info, 0);
OUT_VLAN_ERROR:
    p_item->src_mac_entry = ~0;
    ifx_ppa_drv_del_mac_entry( &route.src_mac, 0);
NEW_SRC_MAC_ERROR:
    p_item->mtu_entry = ~0;
    ifx_ppa_drv_del_mtu_entry(&route.mtu_info, 0);
MTU_ERROR:
    p_item->pppoe_entry = ~0;    
    ifx_ppa_drv_del_pppoe_entry( &route.pppoe_info, 0);
SESSION_VALID_PPPOE_ERROR:
    global_fail_count++;
    return ret;
}

/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/

int32_t ppa_hw_update_session_extra(struct session_list_item *p_item, uint32_t flags)
{    
    PPE_ROUTING_INFO route={0};

    route.mtu_info.mtu_ix = ~0;
    route.out_vlan_info.vlan_entry = ~0;

    if ( (flags & PPA_F_SESSION_NEW_DSCP) )
        route.update_flags |= IFX_PPA_UPDATE_ROUTING_ENTRY_NEW_DSCP_EN | IFX_PPA_UPDATE_ROUTING_ENTRY_NEW_DSCP;

    if ( (flags & PPA_F_SESSION_VLAN) )
        route.update_flags |= IFX_PPA_UPDATE_ROUTING_ENTRY_VLAN_INS_EN | IFX_PPA_UPDATE_ROUTING_ENTRY_NEW_VCI | IFX_PPA_UPDATE_ROUTING_ENTRY_VLAN_RM_EN;

    if ( (flags & PPA_F_MTU) )
    {
        route.mtu_info.mtu_ix = p_item->mtu_entry;
        if ( ifx_ppa_drv_get_mtu_entry( &route.mtu_info, 0) == IFX_SUCCESS )
        {
            if ( route.mtu_info.mtu == p_item->mtu )
            {
                //  entry not changed
                route.mtu_info.mtu_ix = p_item->mtu_entry;
                goto PPA_HW_UPDATE_SESSION_EXTRA_MTU_GOON;
            }
            else
            {
                //  entry changed, so delete old first and create new one later
                ifx_ppa_drv_del_mtu_entry( &route.mtu_info, 0);
                p_item->mtu_entry = ~0;
            }
        }
        //  create new MTU entry
        route.mtu_info.mtu = p_item->mtu;
        if ( ifx_ppa_drv_add_mtu_entry( &route.mtu_info, 0) == IFX_SUCCESS )
        {
            //  success
            p_item->mtu_entry = route.mtu_info.mtu_ix;
            route.update_flags |= IFX_PPA_UPDATE_ROUTING_ENTRY_MTU_IX;
        }
        else
            return IFX_EAGAIN;
    }
PPA_HW_UPDATE_SESSION_EXTRA_MTU_GOON:

    if ( (flags & PPA_F_SESSION_OUT_VLAN) )
    {
        route.out_vlan_info.vlan_entry = p_item->out_vlan_entry;
        if ( ifx_ppa_drv_get_outer_vlan_entry(&route.out_vlan_info, 0) == IFX_SUCCESS )
        {
            if ( route.out_vlan_info.vlan_id == p_item->out_vlan_tag )
            {
                //  entry not changed
                goto PPA_HW_UPDATE_SESSION_EXTRA_OUT_VLAN_GOON;
            }
            else
            {
                //  entry changed, so delete old first and create new one later                
                ifx_ppa_drv_del_outer_vlan_entry(&route.out_vlan_info, 0);
                p_item->out_vlan_entry = ~0;
            }
        }
        //  create new OUT VLAN entry
        route.out_vlan_info.vlan_id = p_item->out_vlan_tag;        
        if ( ifx_ppa_drv_add_outer_vlan_entry( &route.out_vlan_info, 0) == IFX_SUCCESS )
        {
            p_item->out_vlan_entry = route.out_vlan_info.vlan_entry;
            route.update_flags |= IFX_PPA_UPDATE_ROUTING_ENTRY_OUT_VLAN_IX;
        }
        else
            return IFX_EAGAIN;

        route.update_flags |= IFX_PPA_UPDATE_ROUTING_ENTRY_OUT_VLAN_INS_EN | IFX_PPA_UPDATE_ROUTING_ENTRY_OUT_VLAN_RM_EN;
    }
PPA_HW_UPDATE_SESSION_EXTRA_OUT_VLAN_GOON:

    route.entry = p_item->routing_entry;
    route.f_new_dscp_enable = p_item->flags & SESSION_VALID_NEW_DSCP;
    if ( (p_item->flags & SESSION_VALID_NEW_DSCP) )
        route.new_dscp = p_item->new_dscp;
    
    route.f_vlan_ins_enable =p_item->flags & SESSION_VALID_VLAN_INS;
    route.new_vci = p_item->new_vci;
    
    route.f_vlan_rm_enable = p_item->flags & SESSION_VALID_VLAN_RM;

    route.f_out_vlan_ins_enable =p_item->flags & SESSION_VALID_OUT_VLAN_INS;
    route.out_vlan_info.vlan_entry = p_item->out_vlan_entry,    
    
    route.f_out_vlan_rm_enable = p_item->flags & SESSION_VALID_OUT_VLAN_RM;
    
    
    ifx_ppa_drv_update_routing_entry(&route, 0);
    return IFX_SUCCESS;
}

void ppa_hw_del_session(struct session_list_item *p_item)
{
    PPE_ROUTING_INFO route={0};
    
    if ( (p_item->flags & SESSION_ADDED_IN_HW) )
    {
        //  when init, these entry values are ~0, the max the number which can be detected by these functions
        route.entry = p_item->routing_entry;
        global_session_count--;
        ifx_ppa_drv_del_routing_entry(&route, 0);
        p_item->routing_entry = ~0;

        route.out_vlan_info.vlan_entry = p_item->out_vlan_entry;
        ifx_ppa_drv_del_outer_vlan_entry(&route.out_vlan_info, 0);
        p_item->out_vlan_entry = ~0;

        route.pppoe_info.pppoe_ix = p_item->pppoe_entry;
        ifx_ppa_drv_del_pppoe_entry(&route.pppoe_info, 0);
        p_item->pppoe_entry = ~0;

        route.mtu_info.mtu_ix = p_item->mtu_entry;
        ifx_ppa_drv_del_mtu_entry( &route.mtu_info, 0);
        p_item->mtu_entry = ~0;

        route.src_mac.mac_ix = p_item->src_mac_entry;
        ifx_ppa_drv_del_mac_entry( &route.src_mac, 0);
        p_item->src_mac_entry = ~0;

        route.tnnl_info.tunnel_idx = p_item->tunnel_idx;
        if(p_item->flags & SESSION_TUNNEL_6RD){
            route.tnnl_info.tunnel_type = SESSION_TUNNEL_6RD;
        }
        ifx_ppa_drv_del_tunnel_entry(&route.tnnl_info, 0);
        p_item->flags &= ~SESSION_ADDED_IN_HW;
    }
}
 
/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/


/*
 *  multicast routing operation
 */

int32_t ppa_lookup_mc_group(IP_ADDR_C ip_mc_group, struct mc_group_list_item **pp_item)
{
    uint32_t ret;
    struct mc_group_list_item *p_prev, *p_cur;
    struct mc_group_list_item **pp_header;

    ASSERT(pp_item != NULL, "pp_item == NULL");

    ret = IFX_PPA_SESSION_NOT_ADDED;

    pp_header = g_mc_group_list_hash_table + SESSION_LIST_MC_HASH_VALUE(ip_mc_group.ip.ip);
    p_prev = NULL;
    ppa_lock_get(&g_mc_group_list_lock);
    p_cur = *pp_header;
    while ( p_cur )
    {
        if ( ip_mc_group.f_ipv6 == 0 ) 
        {   /*ipv4 address compare */
            if ( p_cur->ip_mc_group.ip.ip == ip_mc_group.ip.ip )
                break;
        }
        else
         { /*ipv6 compare */
           if ( ppa_memcmp( p_cur->ip_mc_group.ip.ip6, ip_mc_group.ip.ip6, sizeof(ip_mc_group.ip.ip6 )  ) == 0 )
            
                break;
         }

        p_prev = p_cur;
        p_cur = p_cur->next;
    }
    if ( p_cur )
    {
        if ( p_prev )
        {
            p_prev->next = p_cur->next;
            p_cur->next = *pp_header;
            *pp_header = p_cur;
        }

        ret = IFX_PPA_SESSION_EXISTS;
    }
    ppa_lock_release(&g_mc_group_list_lock);

    *pp_item = p_cur;

    return ret;
}

/*
    ppa_mc_group_checking: check whether it is valid acceleration session. the result value :
    1) if not found any valid downstream interface, includes num_ifs is zero: return IFX_FAILURE
    2) if downstream interface's VLAN tag not complete same: return IFX_FAILURE
    3) if p_item is NULL: return IFX_ENOMEM;
    
    
*/
int32_t ppa_mc_group_checking(PPA_MC_GROUP *p_mc_group, struct mc_group_list_item *p_item, uint32_t flags)
{
    struct netif_info *p_netif_info;
    uint32_t bit;
    uint32_t idx;
    uint32_t i, bfAccelerate=1, tmp_flag = 0, tmp_out_vlan_tag=0;
    uint16_t  tmp_new_vci=0, bfFirst = 1 ;
    
    if ( !p_item )
        return IFX_ENOMEM;

    //before updating p_item, need to clear some previous values, but cannot memset all p_item esp for p_item's next pointer.
    p_item->dest_ifid = 0;
    //p_item->flags = 0;  //don't simple clear original flag value, esp for flag "SESSION_ADDED_IN_HW"
    p_item->if_mask = 0;
    p_item->new_dscp = 0;
    
    p_item->bridging_flag = p_mc_group->bridging_flag;

    for ( i = 0, bit = 1, idx = 0; i < PPA_MAX_MC_IFS_NUM && idx < p_mc_group->num_ifs; i++, bit <<= 1 )
    {
        if ( p_mc_group->if_mask & bit)
        {
            if ( ppa_netif_lookup(p_mc_group->array_mem_ifs[i].ifname, &p_netif_info) == IFX_SUCCESS )
            {
                //  dest interface
                if ( ppa_netif_update(NULL, p_mc_group->array_mem_ifs[i].ifname) != IFX_SUCCESS
                    || !(p_netif_info->flags & NETIF_PHYS_PORT_GOT) )
                {
                    ppa_netif_put(p_netif_info);
                    ppa_debug(DBG_ENABLE_MASK_DUMP_MC_GROUP, "Warning: No PHYS found for interface %s", p_mc_group->array_mem_ifs[i].ifname);
                    bfAccelerate = 0;
                    break;
                }
                if ( bfFirst )
                {  /* keep the first interface's flag. Make sure all interface's vlan action should be same, otherwise PPE FW cannot do it */
                    tmp_flag = p_netif_info->flags;
                    tmp_new_vci = p_netif_info->inner_vid;
                    tmp_out_vlan_tag = p_netif_info->outer_vid;
                    bfFirst = 0;
                }
                else
                {
                    if ( ( tmp_flag & ( NETIF_VLAN_OUTER | NETIF_VLAN_INNER ) )  != ( p_netif_info->flags & ( NETIF_VLAN_OUTER | NETIF_VLAN_INNER ) ) )
                    {
                        bfAccelerate = 0;
                        ppa_debug(DBG_ENABLE_MASK_DUMP_MC_GROUP, "ppa_add_mc_group not same flag (%0x_%0x)\n", tmp_flag & (NETIF_VLAN_OUTER | NETIF_VLAN_INNER ), p_netif_info->flags & (NETIF_VLAN_OUTER | NETIF_VLAN_INNER) );
                        ppa_netif_put(p_netif_info);
                        break;
                    }
                    else if ( tmp_out_vlan_tag != p_netif_info->outer_vid )
                    {
                        bfAccelerate = 0;
                        ppa_debug(DBG_ENABLE_MASK_DUMP_MC_GROUP, "ppa_add_mc_group not same out vlan tag (%0x_%0x)\n", tmp_out_vlan_tag, p_netif_info->outer_vid);
                        ppa_netif_put(p_netif_info);
                        break;
                    }
                    else if ( tmp_new_vci != p_netif_info->inner_vid )
                    {
                        bfAccelerate = 0;
                        ppa_debug(DBG_ENABLE_MASK_DUMP_MC_GROUP, "ppa_add_mc_group not same inner vlan (%0x_%0x)\n", tmp_new_vci , p_netif_info->inner_vid);
                        ppa_netif_put(p_netif_info);
                        break;
                    }
                }

                p_item->dest_ifid |= 1 << p_netif_info->phys_port;  //  sgh change xuliang's original architecture, but for unicast routing/bridging, still keep old one

                ASSERT(p_netif_info->flags & NETIF_MAC_ENTRY_CREATED, "ETH physical interface must have MAC address");
                p_item->src_mac_entry = p_netif_info->mac_entry;
                if ( !p_mc_group->bridging_flag )
                    p_item->flags |= SESSION_VALID_SRC_MAC;
                else //otherwise clear this bit in case it is set beofre calling this API
                {
                    p_item->flags &= ~SESSION_VALID_SRC_MAC;
                }
                
                p_item->netif[idx] = p_netif_info->netif;
                p_item->ttl[idx]   = p_mc_group->array_mem_ifs[i].ttl;
                p_item->if_mask |= 1 << idx;

                ppa_netif_put(p_netif_info);

                idx++;
            }
            else
            {
                bfAccelerate = 0;
                ppa_debug(DBG_ENABLE_MASK_DUMP_MC_GROUP, "ppa_add_mc_group cannot find the interface:%s)\n", p_mc_group->array_mem_ifs[i].ifname);
                break;
            }
        }
    }

    if ( bfAccelerate == 0 || idx == 0 || (!p_mc_group->bridging_flag && !(p_item->flags & SESSION_VALID_SRC_MAC)) )
    {        
        return IFX_FAILURE;
    }

    //  VLAN
    ASSERT(tmp_flag & NETIF_VLAN_CANT_SUPPORT, "MC processing can support two layers of VLAN only");
    if ( (tmp_flag & NETIF_VLAN_OUTER) )
    {
        p_item->out_vlan_tag = tmp_out_vlan_tag;
        p_item->flags |= SESSION_VALID_OUT_VLAN_INS;
        ppa_debug(DBG_ENABLE_MASK_DUMP_MC_GROUP, "set SESSION_VALID_OUT_VLAN_INS:%x\n", p_item->out_vlan_tag, tmp_new_vci);
    }
    else //otherwise clear this bit in case it is set beofre calling this API
    {
        p_item->flags &= ~SESSION_VALID_OUT_VLAN_INS;
        ppa_debug(DBG_ENABLE_MASK_DUMP_MC_GROUP, "unset SESSION_VALID_OUT_VLAN_INS\n");
    }
    
    if ( (tmp_flag & NETIF_VLAN_INNER) )
    {
        p_item->new_vci = tmp_new_vci;

        p_item->flags |= SESSION_VALID_VLAN_INS;
        ppa_debug(DBG_ENABLE_MASK_DUMP_MC_GROUP, "set SESSION_VALID_VLAN_INS:%x\n", p_item->new_vci);
    }
    else //otherwise clear this bit in case it is set beofre calling this API
    {
        p_item->flags &= ~SESSION_VALID_VLAN_INS;
        ppa_debug(DBG_ENABLE_MASK_DUMP_MC_GROUP, "unset SESSION_VALID_VLAN_INS\n");
    }

   //PPPOE
    if ( !p_mc_group->bridging_flag )
        p_item->flags |= SESSION_VALID_PPPOE;   //  firmware will remove PPPoE header, if and only if the PPPoE header available
    else //otherwise clear this bit in case it is set beofre calling this API
    {
        p_item->flags &= ~SESSION_VALID_PPPOE;
    }

    //  multicast  address
    ppa_memcpy( &p_item->ip_mc_group, &p_mc_group->ip_mc_group, sizeof(p_mc_group->ip_mc_group ) ) ;

    if ( p_mc_group->src_ifname && ppa_netif_lookup(p_mc_group->src_ifname, &p_netif_info) == IFX_SUCCESS )
    {
        //  src interface

        if ( ppa_netif_update(NULL, p_mc_group->src_ifname) == IFX_SUCCESS
            && (p_netif_info->flags & NETIF_PHYS_PORT_GOT) )
        {
            //  PPPoE
            if ( !p_mc_group->bridging_flag && (p_netif_info->flags & NETIF_PPPOE) )
                p_item->flags |= SESSION_VALID_PPPOE;
            else //otherwise clear this bit in case it is set beofre calling this API
            {
                p_item->flags &= ~SESSION_VALID_PPPOE;
            }

            //  VLAN
            ASSERT(p_netif_info->flags & NETIF_VLAN_CANT_SUPPORT, "MC processing can support two layers of VLAN only");
            if ( (p_netif_info->flags & NETIF_VLAN_OUTER) )
                p_item->flags |= SESSION_VALID_OUT_VLAN_RM;
            else //otherwise clear this bit in case it is set beofre calling this API
            {
                p_item->flags &= ~SESSION_VALID_OUT_VLAN_RM;
            }
            
            if ( (p_netif_info->flags & NETIF_VLAN_INNER) )
                p_item->flags |= SESSION_VALID_VLAN_RM;
             else //otherwise clear this bit in case it is set beofre calling this API
            {
                p_item->flags &= ~SESSION_VALID_VLAN_RM;
            }
             
        }
        else  /*not allowed to support non-physical interfaces,like bridge */
        {
            ppa_netif_put(p_netif_info);
            return IFX_FAILURE;
        }
        p_item->src_netif = p_netif_info->netif;

        ppa_netif_put(p_netif_info);
    }

    p_item->num_ifs = idx;

    p_item->dslwan_qid = p_mc_group->dslwan_qid;

    //force update some status by hook itself
    if ( (flags & PPA_F_SESSION_VLAN) )
    {
        if ( p_mc_group->vlan_insert )
        {
            p_item->flags |= SESSION_VALID_VLAN_INS;
            p_item->new_vci = ((uint32_t)p_mc_group->vlan_prio << 13) | ((uint32_t)p_mc_group->vlan_cfi << 12) | (uint32_t)p_mc_group->vlan_id;
        }
        else
        {
            p_item->flags &= ~SESSION_VALID_VLAN_INS;
            p_item->new_vci = 0;
        }

        if ( p_mc_group->vlan_remove )
            p_item->flags |= SESSION_VALID_VLAN_RM;
        else
            p_item->flags &= ~SESSION_VALID_VLAN_RM;
    }     

    if ( (flags & PPA_F_SESSION_OUT_VLAN) )
    {
        if ( p_mc_group->out_vlan_insert )
        {
            p_item->flags |= SESSION_VALID_OUT_VLAN_INS;
            p_item->out_vlan_tag = p_mc_group->out_vlan_tag;
        }
        else
        {
            p_item->flags &= ~SESSION_VALID_OUT_VLAN_INS;
            p_item->out_vlan_tag = 0;
        }

        if ( p_mc_group->out_vlan_remove )
            p_item->flags |= SESSION_VALID_OUT_VLAN_RM;
        else
            p_item->flags &= ~SESSION_VALID_OUT_VLAN_RM;
    }
   
    if ( (flags & PPA_F_SESSION_NEW_DSCP) )
    {
        if ( p_mc_group->new_dscp_en )
        {
            p_item->new_dscp = p_mc_group->new_dscp;
            p_item->flags |= SESSION_VALID_NEW_DSCP;
        }
        else
            p_item->new_dscp &= ~SESSION_VALID_NEW_DSCP;
    }
    
    ppa_debug(DBG_ENABLE_MASK_DUMP_MC_GROUP, "mc flag:%x\n", p_item->flags);
    return IFX_SUCCESS;
}    

int32_t ppa_add_mc_group(PPA_MC_GROUP *p_mc_group, struct mc_group_list_item **pp_item, uint32_t flags)
{
    struct mc_group_list_item *p_item;

    p_item = ppa_alloc_mc_group_list_item();
    if ( !p_item )
        return IFX_ENOMEM;
     
    if( ppa_mc_group_checking(p_mc_group, p_item, flags ) !=IFX_SUCCESS )
    {
        ppa_free_mc_group_list_item(p_item);
        return IFX_FAILURE;
    }
    
    ppa_insert_mc_group_item(p_item);

    *pp_item = p_item;

    return IFX_SUCCESS;
}

int32_t ppa_update_mc_group_extra(PPA_SESSION_EXTRA *p_extra, struct mc_group_list_item *p_item, uint32_t flags)
{
    if ( (flags & PPA_F_SESSION_NEW_DSCP) )
    {
        if ( p_extra->dscp_remark )
        {
            p_item->flags |= SESSION_VALID_NEW_DSCP;
            p_item->new_dscp = p_extra->new_dscp;
        }
        else
            p_item->flags &= ~SESSION_VALID_NEW_DSCP;
    }

    if ( (flags & PPA_F_SESSION_VLAN) )
    {
        if ( p_extra->vlan_insert )
        {
            p_item->flags |= SESSION_VALID_VLAN_INS;
            p_item->new_vci = ((uint32_t)p_extra->vlan_prio << 13) | ((uint32_t)p_extra->vlan_cfi << 12) | p_extra->vlan_id;
        }
        else
        {
            p_item->flags &= ~SESSION_VALID_VLAN_INS;
            p_item->new_vci = 0;
        }

        if ( p_extra->vlan_remove )
            p_item->flags |= SESSION_VALID_VLAN_RM;
        else
            p_item->flags &= ~SESSION_VALID_VLAN_RM;
    }

    if ( (flags & PPA_F_SESSION_OUT_VLAN) )
    {
        if ( p_extra->out_vlan_insert )
        {
            p_item->flags |= SESSION_VALID_OUT_VLAN_INS;
            p_item->out_vlan_tag = p_extra->out_vlan_tag;
        }
        else
        {
            p_item->flags &= ~SESSION_VALID_OUT_VLAN_INS;
            p_item->out_vlan_tag = 0;
        }

        if ( p_extra->out_vlan_remove )
            p_item->flags |= SESSION_VALID_OUT_VLAN_RM;
        else
            p_item->flags &= ~SESSION_VALID_OUT_VLAN_RM;
    }

    if ( p_extra->dslwan_qid_remark )
        p_item->dslwan_qid = p_extra->dslwan_qid;

    return IFX_SUCCESS;
}

void ppa_remove_mc_group(struct mc_group_list_item *p_item)
{
    ppa_debug(DBG_ENABLE_MASK_DUMP_MC_GROUP, "ppa_remove_mc_group:  remove %d from PPA\n", p_item->mc_entry);
    ppa_remove_mc_group_item(p_item);

    ppa_free_mc_group_list_item(p_item);
}

int32_t ppa_mc_group_start_iteration(uint32_t *ppos, struct mc_group_list_item **pp_item)
{
    struct mc_group_list_item *p = NULL;
    int idx;
    uint32_t l;

    l = *ppos + 1;

    ppa_lock_get(&g_mc_group_list_lock);

    for ( idx = 0; l && idx < NUM_ENTITY(g_mc_group_list_hash_table); idx++ )
    {
        for ( p = g_mc_group_list_hash_table[idx]; p; p = p->next )
            if ( !--l )
                break;
    }

    if ( l == 0 && p )
    {
        ++*ppos;
        *pp_item = p;
        return IFX_SUCCESS;
    }
    else
    {
        *pp_item = NULL;
        return IFX_FAILURE;
    }
}

int32_t ppa_mc_group_iterate_next(uint32_t *ppos, struct mc_group_list_item **pp_item)
{
    uint32_t idx;

    if ( *pp_item == NULL )
        return IFX_FAILURE;

    if ( (*pp_item)->next != NULL )
    {
        ++*ppos;
        *pp_item = (*pp_item)->next;
        return IFX_SUCCESS;
    }
    else
    {
        for ( idx = SESSION_LIST_MC_HASH_VALUE((*pp_item)->ip_mc_group.ip.ip) + 1;
              idx < NUM_ENTITY(g_mc_group_list_hash_table);
              idx++ )
            if ( g_mc_group_list_hash_table[idx] != NULL )
            {
                ++*ppos;
                *pp_item = g_mc_group_list_hash_table[idx];
                return IFX_SUCCESS;
            }
        *pp_item = NULL;
        return IFX_FAILURE;
    }
}

void ppa_mc_group_stop_iteration(void)
{
    ppa_lock_release(&g_mc_group_list_lock);
}

/*
 *  multicast routing hardware/firmware operation
 */

int32_t ppa_hw_add_mc_group(struct mc_group_list_item *p_item)
{
    PPE_MC_INFO mc={0};

    mc.out_vlan_info.vlan_entry = ~0;    
    mc.route_type = p_item->bridging_flag ? IFX_PPA_ROUTE_TYPE_NULL : IFX_PPA_ROUTE_TYPE_IPV4;

    //  must be LAN port
    //dest_list = 1 << p_item->dest_ifid;   //  sgh remove it since it is already shifted already
    mc.dest_list = p_item->dest_ifid;          //  due to multiple destination support, the dest_ifid here is a bitmap of destination rather than ifid

    if ( (p_item->flags & SESSION_VALID_OUT_VLAN_INS) )
    {
        mc.out_vlan_info.vlan_id = p_item->out_vlan_tag;
        if ( ifx_ppa_drv_add_outer_vlan_entry( &mc.out_vlan_info, 0) == 0 )
        {
            p_item->out_vlan_entry = mc.out_vlan_info.vlan_entry;
        }
        else
        {
            dbg("add out_vlan_ix error");
            goto OUT_VLAN_ERROR;
        }
    }
    //So far it only support IPv4 yet
    mc.dest_ip_compare = p_item->ip_mc_group.ip.ip;
    
    mc.f_vlan_ins_enable =p_item->flags & SESSION_VALID_VLAN_INS;
    mc.new_vci = p_item->new_vci;
    mc.f_vlan_rm_enable = p_item->flags & SESSION_VALID_VLAN_RM;
    mc.f_src_mac_enable = p_item->flags & SESSION_VALID_SRC_MAC;
    mc.src_mac_ix = p_item->src_mac_entry;
    mc.pppoe_mode = p_item->flags & SESSION_VALID_PPPOE;
    mc.f_out_vlan_ins_enable =p_item->flags & SESSION_VALID_OUT_VLAN_INS;
    mc.out_vlan_info.vlan_entry = p_item->out_vlan_entry;
    mc.f_out_vlan_rm_enable =  p_item->flags & SESSION_VALID_OUT_VLAN_RM;
    mc.f_new_dscp_enable = p_item->flags & SESSION_VALID_NEW_DSCP;
    mc.new_dscp = p_item->new_dscp;
    mc.dest_qid = p_item->dslwan_qid;

    if ( ifx_ppa_drv_add_wan_mc_entry( &mc, 0) == IFX_SUCCESS )
    {
        p_item->mc_entry = mc.p_entry;
        p_item->flags |= SESSION_ADDED_IN_HW;
        return IFX_SUCCESS;
    }

    p_item->out_vlan_entry = ~0;
    ifx_ppa_drv_del_outer_vlan_entry(&mc.out_vlan_info, 0);
OUT_VLAN_ERROR:
    return IFX_EAGAIN;
}

int32_t ppa_hw_update_mc_group_extra(struct mc_group_list_item *p_item, uint32_t flags)
{
    uint32_t update_flags = 0;
    PPE_MC_INFO mc={0};

    if ( (flags & PPA_F_SESSION_NEW_DSCP) )
        update_flags |= IFX_PPA_UPDATE_WAN_MC_ENTRY_NEW_DSCP_EN | IFX_PPA_UPDATE_WAN_MC_ENTRY_NEW_DSCP;

    if ( (flags & PPA_F_SESSION_VLAN) )
        update_flags |=IFX_PPA_UPDATE_WAN_MC_ENTRY_VLAN_INS_EN |IFX_PPA_UPDATE_WAN_MC_ENTRY_NEW_VCI | IFX_PPA_UPDATE_WAN_MC_ENTRY_VLAN_RM_EN;

    if ( (flags & PPA_F_SESSION_OUT_VLAN) )
    {
        mc.out_vlan_info.vlan_entry = p_item->out_vlan_entry;
        if ( ifx_ppa_drv_get_outer_vlan_entry( &mc.out_vlan_info , 0) == IFX_SUCCESS )
        {            
            if ( mc.out_vlan_info.vlan_id == p_item->out_vlan_tag )
            {
                //  entry not changed
                goto PPA_HW_UPDATE_MC_GROUP_EXTRA_OUT_VLAN_GOON;
            }
            else
            {
                //  entry changed, so delete old first and create new one later
                ifx_ppa_drv_del_outer_vlan_entry(&mc.out_vlan_info, 0);
                p_item->out_vlan_entry = ~0;
            }
        }
        //  create new OUT VLAN entry
        mc.out_vlan_info.vlan_id = p_item->out_vlan_tag;
        if ( ifx_ppa_drv_add_outer_vlan_entry( &mc.out_vlan_info, 0) == IFX_SUCCESS )
        {
            p_item->out_vlan_entry = mc.out_vlan_info.vlan_entry;
            update_flags |= IFX_PPA_UPDATE_WAN_MC_ENTRY_OUT_VLAN_IX;
        }
        else
        {
            err("add_outer_vlan_entry fail");
            return IFX_EAGAIN;
        }

        update_flags |= IFX_PPA_UPDATE_WAN_MC_ENTRY_OUT_VLAN_INS_EN | IFX_PPA_UPDATE_WAN_MC_ENTRY_OUT_VLAN_RM_EN ; //IFX_PPA_UPDATE_ROUTING_ENTRY_OUT_VLAN_INS_EN | IFX_PPA_UPDATE_ROUTING_ENTRY_OUT_VLAN_RM_EN;
    }
PPA_HW_UPDATE_MC_GROUP_EXTRA_OUT_VLAN_GOON:
    update_flags |= IFX_PPA_UPDATE_WAN_MC_ENTRY_DEST_QID;  //sgh chnage to update qid, since there is no such flag defined at present

    mc.p_entry = p_item->mc_entry;
    mc.f_vlan_ins_enable = p_item->flags & SESSION_VALID_VLAN_INS;
    mc.new_vci = p_item->new_vci;
    mc.f_vlan_rm_enable = p_item->flags & SESSION_VALID_VLAN_RM;
    mc.f_src_mac_enable = p_item->flags & SESSION_VALID_SRC_MAC;
    mc.src_mac_ix = p_item->src_mac_entry;
    mc.pppoe_mode = p_item->flags & SESSION_VALID_PPPOE;
    mc.f_out_vlan_ins_enable = p_item->flags & SESSION_VALID_OUT_VLAN_INS;
    mc.out_vlan_info.vlan_entry= p_item->out_vlan_entry;
    mc.f_out_vlan_rm_enable = p_item->flags & SESSION_VALID_OUT_VLAN_RM;
    mc.f_new_dscp_enable= p_item->flags & SESSION_VALID_NEW_DSCP;
    mc.new_dscp = p_item->new_dscp;
    mc.dest_qid = p_item->dslwan_qid;
    mc.dest_list = 0;
    mc.update_flags= update_flags;
    ifx_ppa_drv_update_wan_mc_entry(&mc, 0);

    return IFX_SUCCESS;
}

void ppa_hw_del_mc_group(struct mc_group_list_item *p_item)
{
    if ( (p_item->flags & SESSION_ADDED_IN_HW) )
    {
        PPE_MC_INFO mc;
        PPE_OUT_VLAN_INFO out_vlan={0};

        mc.p_entry = p_item->mc_entry;
        out_vlan.vlan_entry = p_item->out_vlan_entry;
        //  when init, these entry values are ~0, the max the number which can be detected by these functions
        ppa_debug(DBG_ENABLE_MASK_DUMP_MC_GROUP, "ppa_hw_del_mc_group:  remove %d from HW\n", p_item->mc_entry);
        ifx_ppa_drv_del_wan_mc_entry(&mc, 0);
        p_item->mc_entry = ~0;

        //  taken from netif_info, so don't need to be removed from MAC table
        p_item->src_mac_entry = ~0;
        
        ifx_ppa_drv_del_outer_vlan_entry( &out_vlan, 0);
        p_item->out_vlan_entry = ~0;

        p_item->flags &= ~SESSION_ADDED_IN_HW;
    }
}

/*
 *  routing polling timer
 */

void ppa_set_polling_timer(uint32_t polling_time, uint32_t force)
{
    if ( polling_time < g_hit_polling_time )
    {
        //  remove timer
        ppa_timer_del(&g_hit_stat_timer);

        //  timeout can not be zero
        g_hit_polling_time = polling_time < 1 ? 1 : polling_time;

        //  check hit stat in case the left time is less then the new timeout
        ppa_check_hit_stat(0);  //  timer is added in this function
    }
    else if ( (polling_time > g_hit_polling_time) && force )
    {
        g_hit_polling_time = polling_time;
    }
}

/*
 *  bridging session operation
 */

int32_t ppa_bridging_lookup_session(uint8_t *mac, PPA_NETIF *netif, struct bridging_session_list_item **pp_item)
{
    int32_t ret;
    struct bridging_session_list_item *p_prev, *p_cur;
    struct bridging_session_list_item **pp_header;

    ASSERT(pp_item != NULL, "pp_item == NULL");

    ret = IFX_PPA_SESSION_NOT_ADDED;

    pp_header = g_bridging_session_list_hash_table + BRIDGING_SESSION_LIST_HASH_VALUE(mac);
    p_prev = NULL;
    ppa_lock_get(&g_bridging_session_list_lock);
    p_cur = *pp_header;
    while ( p_cur && ppa_memcmp(mac, p_cur->mac, PPA_ETH_ALEN) != 0 /* !(*(uint32_t *)mac == *(uint32_t *)p_cur->mac && *((uint16_t *)mac + 2) == *((uint16_t *)p_cur->mac + 2)) */ )
    {
        p_prev = p_cur;
        p_cur = p_cur->next;
    }
    if ( p_cur )
    {
        if ( p_prev )
        {
            p_prev->next = p_cur->next;
            p_cur->next = *pp_header;
            *pp_header = p_cur;
        }

        ret = IFX_PPA_SESSION_EXISTS;
    }
    ppa_lock_release(&g_bridging_session_list_lock);

    *pp_item = p_cur;

    return ret;
}

int32_t ppa_bridging_add_session(uint8_t *mac, PPA_NETIF *netif, struct bridging_session_list_item **pp_item, uint32_t flags)
{
    struct bridging_session_list_item *p_item;
    struct netif_info *ifinfo;

    if ( ppa_netif_update(netif, NULL) != IFX_SUCCESS )
        return IFX_ENOTPOSSIBLE;

    if ( ppa_netif_lookup(ppa_get_netif_name(netif), &ifinfo) != IFX_SUCCESS )
        return IFX_FAILURE;

#if defined(CONFIG_IFX_PPA_API_DIRECTPATH_BRIDGING)  
    if( ifx_ppa_get_ifid_for_netif(netif) > 0 ) return IFX_FAILURE;   // no need to learn and program mac address in ppe/switch if directpath bridging feature is disabled
#endif

    p_item = ppa_bridging_alloc_session_list_item();
    if ( !p_item )
    {
        ppa_netif_put(ifinfo);
        return IFX_ENOMEM;
    }

    dump_bridging_list_item(p_item, "ppa_bridging_add_session (after init)");

    ppa_memcpy(p_item->mac, mac, PPA_ETH_ALEN);
    p_item->netif         = netif;
    p_item->timeout       = ppa_bridging_get_default_session_timeout();
    p_item->last_hit_time = ppa_get_time_in_sec();

    //  TODO: vlan related fields

    //  decide destination list
    /*yixin: VR9 need to add mac address to switch only if the source mac not from switch port  */
    if ( !(ifinfo->flags & NETIF_PHYS_PORT_GOT) )
    {
        PPE_COUNT_CFG count={0};
        ifx_ppa_drv_get_number_of_phys_port( &count, 0);
        p_item->dest_ifid =  count.num + 1;   //trick here: it will be used for PPE Driver's HAL     
    }else{
        p_item->dest_ifid = ifinfo->phys_port;
    }
    if ( (ifinfo->flags & NETIF_PHY_ATM) )
        p_item->dslwan_qid = ifinfo->dslwan_qid;

    if ( (flags & PPA_F_STATIC_ENTRY) )
    {
        p_item->flags    |= SESSION_STATIC;
        p_item->timeout   = ~0; //  max timeout
    }

    if ( (flags & PPA_F_DROP_PACKET) )
        p_item->flags    |= SESSION_DROP;

    ppa_bridging_insert_session_item(p_item);

    dump_bridging_list_item(p_item, "ppa_bridging_add_session (after setting)");

    *pp_item = p_item;

    ppa_netif_put(ifinfo);

    return IFX_SUCCESS;
}

void ppa_bridging_remove_session(struct bridging_session_list_item *p_item)
{
    ppa_bridging_remove_session_item(p_item);

    ppa_bridging_free_session_list_item(p_item);
}

void dump_bridging_list_item(struct bridging_session_list_item *p_item, char *comment)
{
#if defined(DEBUG_DUMP_LIST_ITEM) && DEBUG_DUMP_LIST_ITEM

    if ( !(g_ppa_dbg_enable & DBG_ENABLE_MASK_DUMP_BRIDGING_SESSION) )
        return;

    if ( comment )
        printk("dump_bridging_list_item - %s\n", comment);
    else
        printk("dump_bridging_list_item\n");
    printk("  next             = %08X\n", (uint32_t)p_item->next);
    printk("  mac[6]           = %02x:%02x:%02x:%02x:%02x:%02x\n", p_item->mac[0], p_item->mac[1], p_item->mac[2], p_item->mac[3], p_item->mac[4], p_item->mac[5]);
    printk("  netif            = %08X\n", (uint32_t)p_item->netif);
    printk("  timeout          = %d\n",   p_item->timeout);
    printk("  last_hit_time    = %d\n",   p_item->last_hit_time);
    printk("  flags            = %08X\n", p_item->flags);
    printk("  bridging_entry   = %08X\n", p_item->bridging_entry);

#endif
}

int32_t ppa_bridging_session_start_iteration(uint32_t *ppos, struct bridging_session_list_item **pp_item)
{
    struct bridging_session_list_item *p = NULL;
    int idx;
    uint32_t l;

    l = *ppos + 1;

    ppa_lock_get(&g_bridging_session_list_lock);

    for ( idx = 0; l && idx < NUM_ENTITY(g_bridging_session_list_hash_table); idx++ )
    {
        for ( p = g_bridging_session_list_hash_table[idx]; p; p = p->next )
            if ( !--l )
                break;
    }

    if ( l == 0 && p )
    {
        ++*ppos;
        *pp_item = p;
        return IFX_SUCCESS;
    }
    else
    {
        *pp_item = NULL;
        return IFX_FAILURE;
    }
}

int32_t ppa_bridging_session_iterate_next(uint32_t *ppos, struct bridging_session_list_item **pp_item)
{
    uint32_t idx;

    if ( *pp_item == NULL )
        return IFX_FAILURE;

    if ( (*pp_item)->next != NULL )
    {
        ++*ppos;
        *pp_item = (*pp_item)->next;
        return IFX_SUCCESS;
    }
    else
    {
        for ( idx = BRIDGING_SESSION_LIST_HASH_VALUE((*pp_item)->mac) + 1;
              idx < NUM_ENTITY(g_bridging_session_list_hash_table);
              idx++ )
            if ( g_bridging_session_list_hash_table[idx] != NULL )
            {
                ++*ppos;
                *pp_item = g_bridging_session_list_hash_table[idx];
                return IFX_SUCCESS;
            }
        *pp_item = NULL;
        return IFX_FAILURE;
    }
}

void ppa_bridging_session_stop_iteration(void)
{
    ppa_lock_release(&g_bridging_session_list_lock);
}

/*
 *  bridging session hardware/firmware operation
 */

int32_t ppa_bridging_hw_add_session(struct bridging_session_list_item *p_item)
{
    PPE_BR_MAC_INFO br_mac={0};

    br_mac.port = p_item->dest_ifid;

    if ( (p_item->flags & SESSION_DROP) )
        br_mac.dest_list = 0;  //  no dest list, dropped
    else
        br_mac.dest_list = 1 << p_item->dest_ifid;

    ppa_memcpy(br_mac.mac, p_item->mac, sizeof(br_mac.mac));
    br_mac.f_src_mac_drop = p_item->flags & SESSION_SRC_MAC_DROP_EN;
    br_mac.dslwan_qid = p_item->dslwan_qid;

    if ( ifx_ppa_drv_add_bridging_entry(&br_mac, 0) == IFX_SUCCESS )
    {
        p_item->bridging_entry = br_mac.p_entry;
        p_item->flags |= SESSION_ADDED_IN_HW;
        dbg("add_brdiging_entry successful");
        return IFX_SUCCESS;
    }
    dbg("add_brdiging_entry failed");
    return IFX_FAILURE;
}

void ppa_bridging_hw_del_session(struct bridging_session_list_item *p_item)
{
    if ( (p_item->flags & SESSION_ADDED_IN_HW) )
    {
        PPE_BR_MAC_INFO br_mac={0};
        br_mac.p_entry = p_item->bridging_entry;
        ifx_ppa_drv_del_bridging_entry(&br_mac, 0);
        p_item->bridging_entry = ~0;

        p_item->flags &= ~SESSION_ADDED_IN_HW;
    }
}

/*
 *  bridging polling timer
 */

void ppa_bridging_set_polling_timer(uint32_t polling_time)
{
    if ( polling_time < g_bridging_hit_polling_time )
    {
        //  remove timer
        ppa_timer_del(&g_bridging_hit_stat_timer);

        //  timeout can not be zero
        g_bridging_hit_polling_time = polling_time < 1 ? 1 : polling_time;

        //  check hit stat in case the left time is less then the new timeout
        ppa_bridging_check_hit_stat(0); //  timer is added in this function
    }
}

/*
 *  special function
 */

void ppa_remove_sessions_on_netif(PPA_IFNAME *ifname, int f_is_lan)
{
    struct netif_info *ifinfo;
    uint32_t lan_wan_flag;      //  bit 0: lan, bit 1: wan

    if ( ppa_netif_lookup(ifname, &ifinfo) == IFX_SUCCESS )
        lan_wan_flag = 3;
    else
        lan_wan_flag = f_is_lan ? 1 : 2;

    ppa_remove_routing_sessions_on_netif(ifname, lan_wan_flag);
    if ( lan_wan_flag == 2 )
        ppa_remove_mc_groups_on_netif(ifname);
    if ( lan_wan_flag == 3 )
        ppa_remove_bridging_sessions_on_netif(ifname);
}


static inline void avm_sessionh_table_remove(unsigned int rout_table_entry, unsigned int is_lan){
    unsigned int index = rout_table_entry & 0x7FFFFFFF;
    RW_LOCK_DECLARE;

    if( is_lan ){
        if (likely(index < avm_sessionh_table_lan_size )){
            LAN_TABLE_WRITE_LOCK();
            avm_sessionh_table_lan[index] = INVALID_SESSION;
            LAN_TABLE_WRITE_UNLOCK();
        }
    }
    else {
        if (likely(index < avm_sessionh_table_wan_size )){
            WAN_TABLE_WRITE_LOCK();
            avm_sessionh_table_wan[index] = INVALID_SESSION;
            WAN_TABLE_WRITE_UNLOCK();
        }
    }
    return;
}

static inline void avm_sessionh_table_set(unsigned int rout_table_entry, avm_session_handle sessionh, unsigned int is_lan, unsigned int is_mc ){
    unsigned int index = rout_table_entry & 0x7FFFFFFF;
    RW_LOCK_DECLARE;

    if ( is_mc )
        return;

    if( is_lan ){
        if (likely(index < avm_sessionh_table_lan_size )){

            LAN_TABLE_WRITE_LOCK();
            avm_sessionh_table_lan[index] = sessionh;
            LAN_TABLE_WRITE_UNLOCK();

        } else {
            printk(KERN_ERR "no space for index %d in lan session table\n", rout_table_entry);
        }
    }
    else {
        if (likely(index < avm_sessionh_table_wan_size )) {

            WAN_TABLE_WRITE_LOCK();
            avm_sessionh_table_wan[index] = sessionh;
            WAN_TABLE_WRITE_UNLOCK();

        } else {

            printk(KERN_ERR "no space for index %d in wan session table\n", rout_table_entry);

        }
    }
    return;
}

avm_session_handle avm_session_handle_fast_lookup( unsigned int rout_table_entry, unsigned int is_lan, unsigned int is_mc ){
    unsigned int index = rout_table_entry & 0x7FFFFFFF;
    RW_LOCK_DECLARE;

    avm_session_handle res = INVALID_SESSION;
    if ( unlikely(is_mc) )
        goto SLOW_LOOKUP;


    if( is_lan ) {
        if (likely(index < avm_sessionh_table_lan_size )){
            LAN_TABLE_READ_LOCK();
            res = avm_sessionh_table_lan[index];
            LAN_TABLE_READ_UNLOCK();
#if defined(CONFIG_DEBUG_AVM_PPA_VDEV)
            avm_session_hit_count_lan_table[index] += 1;
#endif
        }
    }
    else {
        if (likely(index < avm_sessionh_table_wan_size )){

            WAN_TABLE_READ_LOCK();
            res = avm_sessionh_table_wan[index];
            WAN_TABLE_READ_UNLOCK();
#if defined(CONFIG_DEBUG_AVM_PPA_VDEV)
            avm_session_hit_count_wan_table[index] += 1;
#endif
        }
    }
SLOW_LOOKUP:
    if (res == INVALID_SESSION )
        return avm_session_handle_slow_lookup(rout_table_entry, is_lan, is_mc);
    else
        return res;
}


/*
 * ####################################
 *           Init/Cleanup API
 * ####################################
 */

int32_t ppa_api_session_manager_init(void)
{

    PPA_MAX_ENTRY_INFO entry;

    //  start timer
    ppa_timer_init(&g_hit_stat_timer, ppa_check_hit_stat);
    ppa_timer_add(&g_hit_stat_timer, g_hit_polling_time);
    ppa_timer_init(&g_bridging_hit_stat_timer, ppa_bridging_check_hit_stat);
    ppa_timer_add(&g_bridging_hit_stat_timer, g_bridging_hit_polling_time);

    /* setup avm_pa sessionh lookup_table */
	printk(KERN_ERR "[%s] avm_pa sessionh_lookup table\n", __FUNCTION__);
	ifx_ppa_drv_get_max_entries( &entry, 0);

    printk(KERN_ERR "max_lan_entries       %d\n", entry.max_lan_entries       );
    printk(KERN_ERR "max_wan_entries       %d\n", entry.max_wan_entries       );
    printk(KERN_ERR "max_mc_entries        %d\n", entry.max_mc_entries        );
    printk(KERN_ERR "max_bridging_entries  %d\n", entry.max_bridging_entries  );
    printk(KERN_ERR "max_ipv6_addr_entries %d\n", entry.max_ipv6_addr_entries );
    printk(KERN_ERR "max_fw_queue          %d\n", entry.max_fw_queue          );
    printk(KERN_ERR "max_6rd_entries       %d\n", entry.max_6rd_entries       );

    avm_sessionh_table_lan_size = entry.max_lan_entries;
    avm_sessionh_table_wan_size = entry.max_wan_entries;

    avm_sessionh_table_lan = kmalloc( avm_sessionh_table_lan_size * sizeof(avm_session_handle), GFP_KERNEL );
    avm_sessionh_table_wan = kmalloc( avm_sessionh_table_wan_size * sizeof(avm_session_handle), GFP_KERNEL );

    if ( !avm_sessionh_table_lan || !avm_sessionh_table_wan ){
        printk(KERN_ERR "cannot alloc lan/wan-lookup table %d, %d\n", entry.max_lan_entries, entry.max_wan_entries);
        BUG();
    }
    memset( avm_sessionh_table_lan, INVALID_SESSION, avm_sessionh_table_lan_size * sizeof(avm_session_handle) );
    memset( avm_sessionh_table_wan, INVALID_SESSION, avm_sessionh_table_wan_size * sizeof(avm_session_handle) );

#if defined(CONFIG_DEBUG_AVM_PPA_VDEV)
    avm_session_hit_count_lan_table = kmalloc( avm_sessionh_table_lan_size * sizeof(unsigned int), GFP_KERNEL );
    avm_session_hit_count_wan_table = kmalloc( avm_sessionh_table_wan_size * sizeof(unsigned int), GFP_KERNEL );
    if ( !avm_session_hit_count_lan_table || !avm_session_hit_count_lan_table){
        printk(KERN_ERR "cannot alloc hit_count_table table %d, %d\n", entry.max_lan_entries, entry.max_wan_entries);
        BUG();
    }
    memset( avm_session_hit_count_lan_table, 0, avm_sessionh_table_lan_size * sizeof(unsigned int) );
    memset( avm_session_hit_count_wan_table, 0, avm_sessionh_table_wan_size * sizeof(unsigned int) );
#endif



    return IFX_SUCCESS;
}

void ppa_api_session_manager_exit(void)
{
    ppa_timer_del(&g_hit_stat_timer);
    ppa_timer_del(&g_bridging_hit_stat_timer);

    ppa_free_session_list();
    ppa_free_mc_group_list();
    ppa_free_bridging_session_list();

    ppa_kmem_cache_shrink(g_session_item_cache);
    ppa_kmem_cache_shrink(g_mc_group_item_cache);
    ppa_kmem_cache_shrink(g_bridging_session_item_cache);

    /* free avm_pa sessionh lookup_table */
    kfree (avm_sessionh_table_lan);
    avm_sessionh_table_lan = NULL;

    kfree (avm_sessionh_table_wan);
    avm_sessionh_table_wan = NULL;

#if defined(CONFIG_DEBUG_AVM_PPA_VDEV)
    kfree (avm_session_hit_count_lan_table);
    avm_session_hit_count_lan_table = NULL;

    kfree (avm_session_hit_count_wan_table);
    avm_session_hit_count_wan_table = NULL;
#endif

}

int32_t ppa_api_session_manager_create(void)
{
	printk(KERN_ERR "[%s]\n", __FUNCTION__);
    if ( ppa_mem_cache_create("ppa_session_item", sizeof(struct session_list_item), &g_session_item_cache) )
    {
        err("Failed in creating mem cache for routing session list.");
        goto PPA_API_SESSION_MANAGER_CREATE_FAIL;
    }

    if ( ppa_mem_cache_create("mc_group_item", sizeof(struct mc_group_list_item), &g_mc_group_item_cache) )
    {
        err("Failed in creating mem cache for multicast group list.");
        goto PPA_API_SESSION_MANAGER_CREATE_FAIL;
    }

    if ( ppa_mem_cache_create("bridging_sess_item", sizeof(struct bridging_session_list_item), &g_bridging_session_item_cache) )
    {
        err("Failed in creating mem cache for bridging session list.");
        goto PPA_API_SESSION_MANAGER_CREATE_FAIL;
    }

    if ( ppa_lock_init(&g_session_list_lock) )
    {
        err("Failed in creating lock for routing session list.");
        goto PPA_API_SESSION_MANAGER_CREATE_FAIL;
    }

    if ( ppa_lock_init(&g_mc_group_list_lock) )
    {
        err("Failed in creating lock for multicast group list.");
        goto PPA_API_SESSION_MANAGER_CREATE_FAIL;
    }

    if ( ppa_lock_init(&g_bridging_session_list_lock) )
    {
        err("Failed in creating lock for bridging session list.");
        goto PPA_API_SESSION_MANAGER_CREATE_FAIL;
    }

    return IFX_SUCCESS;

PPA_API_SESSION_MANAGER_CREATE_FAIL:
    ppa_api_session_manager_destroy();
    return IFX_EIO;
}

void ppa_api_session_manager_destroy(void)
{
    if ( g_session_item_cache )
    {
        ppa_mem_cache_destroy(g_session_item_cache);
        g_session_item_cache = NULL;
    }

    if ( g_mc_group_item_cache )
    {
        ppa_mem_cache_destroy(g_mc_group_item_cache);
        g_mc_group_item_cache = NULL;
    }

    if ( g_bridging_session_item_cache )
    {
        ppa_mem_cache_destroy(g_bridging_session_item_cache);
        g_bridging_session_item_cache = NULL;
    }

    ppa_lock_destroy(&g_session_list_lock);

    ppa_lock_destroy(&g_mc_group_list_lock);

    ppa_lock_destroy(&g_bridging_session_list_lock);
}

uint32_t ppa_api_get_session_poll_timer(void)
{
    return g_hit_polling_time;
}

#ifdef CONFIG_AVM_PA
/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/


void update_atm_flag(struct avm_pa_pid_hwinfo *info){
	if (info->atmvcc)
		info->flags |= AVMNET_DEVICE_IFXPPA_ATM_WAN;
}

/*------------------------------------------------------------------------------------------*\
 * ADD_SESSION_STATES:
\*------------------------------------------------------------------------------------------*/
#define AASS_INGRESS_IPV4 		(1 << 0)
#define AASS_INGRESS_IPV6 		(1 << 1)
#define AASS_INGRESS_PPPOE 		(1 << 2)
#define AASS_INGRESS_PPPOA 		(1 << 3)
#define AASS_INGRESS_IPOA 		(1 << 4)
#define AASS_INGRESS_ETHERNET 	(1 << 5)

#define AASS_EGRESS_IPV4 		(1 << 6)
#define AASS_EGRESS_IPV6 		(1 << 7)
#define AASS_EGRESS_PPPOE 		(1 << 8)
#define AASS_EGRESS_PPPOA 		(1 << 9)
#define AASS_EGRESS_IPOA 		(1 << 10)
#define AASS_EGRESS_ETHERNET 	(1 << 11)

#define AASS_UPDATE_OLD_SESSION 	(1 << 12)
#define MAX_SUPPORTED_VLAN_TAGS 2
/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
void print_flags(unsigned int flags){

	if ( !flags || !( g_ppa_dbg_enable & DBG_ENABLE_MASK_DEBUG_PRINT) )
		return;

	printk(KERN_ERR "flags: ");
	if (flags & AASS_INGRESS_IPV4          ) printk(" INGRESS_IPV4 "  );
	if (flags & AASS_INGRESS_IPV6          ) printk(" INGRESS_IPV6 "  );
	if (flags & AASS_INGRESS_PPPOE         ) printk(" INGRESS_PPPOE " );
	if (flags & AASS_INGRESS_PPPOA         ) printk(" INGRESS_PPPOA " );
	if (flags & AASS_INGRESS_IPOA 	       ) printk(" INGRESS_IPOA "  );
	if (flags & AASS_INGRESS_ETHERNET      ) printk(" INGRESS_ETHERNET ");
	if (flags & AASS_EGRESS_IPV4 	       ) printk(" EGRESS_IPV4 "   );
	if (flags & AASS_EGRESS_IPV6 	       ) printk(" EGRESS_IPV6 "   );
	if (flags & AASS_EGRESS_PPPOE 	       ) printk(" EGRESS_PPPOE "  );
	if (flags & AASS_EGRESS_PPPOA 	       ) printk(" EGRESS_PPPOA "  );
	if (flags & AASS_EGRESS_IPOA 	       ) printk(" EGRESS_IPOA "   );
	if (flags & AASS_EGRESS_ETHERNET       ) printk(" EGRESS_ETHERNET ");
	if (flags & AASS_UPDATE_OLD_SESSION    ) printk(" UPDATE_OLD_SESSION ");
	printk(KERN_ERR "\n ");
}
/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/

int avm_pa_add_ifx_session( struct avm_pa_session *avm_session ) {

    int aass = 0; //AVM_ADD_SESSION_STATE
    uint16_t incoming_vlan_tci[MAX_SUPPORTED_VLAN_TAGS];
    uint16_t outgoing_vlan_tci[MAX_SUPPORTED_VLAN_TAGS];
    uint16_t nr_vlan_out = 0;
    uint16_t nr_vlan_in = 0;
    struct session_list_item *p_item = NULL;
    int hw_ret;
    int n;
    uint16_t session_id; 
    struct avm_pa_pid_hwinfo *ingress_hw;
    struct avm_pa_pid_hwinfo *egress_hw;

    ppa_debug(DBG_ENABLE_MASK_AVM_PPA, "[%s] \n", __FUNCTION__);

    /* constraints */
    if ( !avm_session ){
        ppa_debug(DBG_ENABLE_MASK_AVM_PPA, "[%s] error: no avm_pa_session\n", __FUNCTION__);
        goto ERR;
    }

    session_id = avm_session->session_handle;

    if ( !session_id ){
        ppa_debug(DBG_ENABLE_MASK_AVM_PPA, "[%s] error: no session_id\n", __FUNCTION__);
        goto ERR;
    }

    if(avm_session->bsession != NULL){
        ppa_debug(DBG_ENABLE_MASK_AVM_PPA, " [%s] Refusing bridging session in IFX PA\n", __func__);
        goto ERR;
    }

    if( avm_session->negress != 1 ) {
        ppa_debug(DBG_ENABLE_MASK_AVM_PPA, "[%s] multiple egress not supported\n", __FUNCTION__);
        goto ERR;
    }

    avm_lookup_session(session_id, &p_item);

    if(p_item){
        aass |= AASS_UPDATE_OLD_SESSION;
        ppa_debug(DBG_ENABLE_MASK_AVM_PPA, "[%s] got old session %d\n", __FUNCTION__, session_id);
        ppa_debug(DBG_ENABLE_MASK_AVM_PPA, "[%s] del old item\n", __FUNCTION__);

        ppa_hw_del_session(p_item);
        ppa_init_session_list_item(p_item);
        ppa_debug(DBG_ENABLE_MASK_AVM_PPA, "[%s] deactivate ppa\n", __FUNCTION__);
        ifx_ppa_pwm_deactivate_module();

    } else {
        p_item = ppa_alloc_session_list_item();
        if ( !p_item ) {
            ppa_debug(DBG_ENABLE_MASK_AVM_PPA, "[%s] failed in memory allocation\n", __FUNCTION__);
            goto ERR;
        }
        ppa_debug(DBG_ENABLE_MASK_AVM_PPA, "[%s] session item allocated, %p\n", __FUNCTION__, p_item);
    }

    p_item->session_id = session_id;

    switch ( avm_session->ingress.pkttype & AVM_PA_PKTTYPE_PROTO_MASK ) {
        case IPPROTO_TCP:
            p_item->flags |= SESSION_IS_TCP; 
            break;
        case IPPROTO_UDP:
            p_item->flags &= ~SESSION_IS_TCP; 
            break;
        default:
            ppa_debug(DBG_ENABLE_MASK_AVM_PPA, "[%s] no support for PKTTYPE != ( UDP | TCP )\n", __FUNCTION__);
            goto ERR_FREE_PITEM;
    }

    ingress_hw = avm_pa_pid_get_hwinfo( avm_session->ingress_pid_handle );
    egress_hw  = avm_pa_pid_get_hwinfo( avm_session->egress[0].pid_handle );

    if (!ingress_hw || !egress_hw ) {

    	ppa_debug(DBG_ENABLE_MASK_AVM_PPA, "hw_pointer not valid: ingress_hw=%p, egress_hw=%p \n", ingress_hw, egress_hw);
    	goto ERR_FREE_PITEM;
    }

    update_atm_flag( ingress_hw );
    update_atm_flag( egress_hw );

    p_item->dest_port_map = 0;


    // OUTGOING VIRTUAL CHANNEL (e.g. WLAN)

    if ( egress_hw->flags & AVMNET_DEVICE_IFXPPA_VIRT_TX ) {
    	if ( disable_wlan_egress_sessions ){
    		ppa_debug(DBG_ENABLE_MASK_AVM_PPA, "wlan egress not enabled, yet\n");
    		goto ERR_FREE_PITEM;
    	}
    	p_item->dest_ifid = egress_hw->virt_tx_dev->hw_dev_id + IFX_VIRT_DEV_OFFSET;
    	p_item->dest_port_map = 0;

    	if ( enable_wlan_dscp ) {
    		p_item->new_dscp = enable_wlan_dscp;
    		p_item->flags |= SESSION_VALID_NEW_DSCP;
    	}
    	if ( enable_wlan_qid ) {
    		p_item->dslwan_qid = enable_wlan_qid;
    		// p_item->flags |= SESSION_VALID_NEW_DSCP;
    	}

        ppa_debug(DBG_ENABLE_MASK_AVM_PPA, "[%s] xxx -> wlan (dest_ifid =%d) \n", __FUNCTION__ , p_item->dest_ifid);
    }


    // OUTGOING ETHERNET

    if ( egress_hw->flags & ( AVMNET_DEVICE_IFXPPA_ETH_LAN | AVMNET_DEVICE_IFXPPA_ETH_WAN ) ) {
    	// XXX -> LAN
    	if ( use_avm_dest_port_map ) {
    		p_item->dest_ifid = 0;
    		p_item->dest_port_map = (1 << egress_hw->mac_nr);
    		p_item->flags  |= SESSION_VALID_AVM_PORTMAP;
    	} else {
    		if ( egress_hw->mac_nr < 2 ){
    			p_item->dest_ifid = egress_hw->mac_nr;
    			p_item->dest_port_map = 0;
        		p_item->flags  &= ~SESSION_VALID_AVM_PORTMAP;
    		} else {
    			goto ERR_FREE_PITEM;
    		}

    	}
        ppa_debug(DBG_ENABLE_MASK_AVM_PPA, "[%s] xxx -> lan/ata (dstmac =%d) \n", __FUNCTION__ , egress_hw->mac_nr);
    }

    // INCOMMING ETHERNET

    if ( (ingress_hw->flags & AVMNET_DEVICE_IFXPPA_ETH_WAN ) && in_avm_ata_mode()) {
   		// ATA -> XXX
    	p_item->flags |= SESSION_WAN_ENTRY;
    } else if ( ingress_hw->flags & ( AVMNET_DEVICE_IFXPPA_ETH_LAN | AVMNET_DEVICE_IFXPPA_VIRT_RX )) {
    	// (LAN | WLAN) -> XXX
    	//TODO ADD WLAN-ATA-Mode-Handling
    	p_item->flags |= SESSION_LAN_ENTRY;
    }

    //  DSL DOWNSTREAM

    if ( ingress_hw->flags & (AVMNET_DEVICE_IFXPPA_PTM_WAN | AVMNET_DEVICE_IFXPPA_ATM_WAN) ) {
    	// (V/A)DSL-WAN -> LAN
		p_item->flags |= SESSION_WAN_ENTRY;
        p_item->flags |= SESSION_VALID_DSLWAN_QID;
        if ( ingress_hw->flags & AVMNET_DEVICE_IFXPPA_ATM_WAN ) {
        	p_item->dslwan_qid = ifx_ppa_drv_get_dslwan_qid_with_vcc( (struct atm_vcc*)ingress_hw->atmvcc );
        } else {
        	p_item->dslwan_qid = 3; //TODO das scheint egal zu sein?!
        }
        ppa_debug(DBG_ENABLE_MASK_AVM_PPA, "[%s] dslwan -> lan: qid = %d \n", __FUNCTION__ , p_item->dslwan_qid);
    }

    //  DSL UPSTREAM

    if ( egress_hw->flags & AVMNET_DEVICE_IFXPPA_ATM_WAN ) {
   		// LAN -> WAN ADSL
    	p_item->flags |= SESSION_LAN_ENTRY | SESSION_VALID_DSLWAN_QID;
        p_item->dest_ifid = 7;
        p_item->dslwan_qid = ifx_ppa_drv_get_dslwan_qid_with_vcc( (struct atm_vcc*)egress_hw->atmvcc );
        ppa_debug(DBG_ENABLE_MASK_AVM_PPA, "[%s] LAN -> ADSL: qid = %d\n", __FUNCTION__,p_item->dslwan_qid );
    }

    if ( egress_hw->flags & AVMNET_DEVICE_IFXPPA_PTM_WAN ) {
   		// LAN -> WAN VDSL
    	p_item->flags |= SESSION_LAN_ENTRY | SESSION_VALID_DSLWAN_QID;
        p_item->dest_ifid = 7;
        p_item->dslwan_qid = 1; // ptm braucht bei uns 1
        ppa_debug(DBG_ENABLE_MASK_AVM_PPA, "[%s] LAN -> VDSL\n", __FUNCTION__);
    }

    if (( p_item->flags & SESSION_LAN_ENTRY & SESSION_WAN_ENTRY ) ||
    		(( p_item->flags & (SESSION_LAN_ENTRY | SESSION_WAN_ENTRY)) == 0) ){
    	err("session cannot be classified");
        goto ERR_FREE_PITEM;
    }

    if (( p_item->flags & SESSION_LAN_ENTRY ) && disable_lan_sessions){
    	err("disable_lan_sessions is set");
        goto ERR_FREE_PITEM;
    }

    ppa_debug(DBG_ENABLE_MASK_AVM_PPA, "[%s] Try to add session with if: ( %d  -> %d)\n", __FUNCTION__, ingress_hw->mac_nr, egress_hw->mac_nr);

    p_item->mtu = avm_session->egress[0].mtu;

     /*--------------------*\
      * Ingress (pkt match)
     \*--------------------*/
    for (n=0; n < avm_session->ingress.nmatch; n++) {
        struct avm_pa_match_info *p = &avm_session->ingress.match[n];
        hdrunion_t *hdr = (hdrunion_t *)&avm_session->ingress.hdrcopy[p->offset + avm_session->ingress.hdroff];
    	int num_supported_ingerss_vlan_tags = (ingress_hw->flags & AVMNET_DEVICE_IFXPPA_ATM_WAN)? MAX_SUPPORTED_VLAN_TAGS : 1;

        switch (p->type) {
            case AVM_PA_ETH:
                memcpy(&p_item->src_mac, &hdr->ethh.h_source, PPA_ETH_ALEN);
                aass |= AASS_INGRESS_ETHERNET;
                break;

            case AVM_PA_IPV4:
                aass |= AASS_INGRESS_IPV4;
                if ( aass & AASS_INGRESS_IPV6 ){
                	// das sollte eigentlich nicht passieren, da erst ipv4 u. danach ipv6 header geparsed wird
                	dbg("Session seems to be 6RD, put IPv4-addresses into tunnel field");
                	p_item->tun_saddr.ip = hdr->iph.saddr;
                	p_item->tun_daddr.ip = hdr->iph.daddr;
                }
                else{
                	p_item->src_ip.ip = hdr->iph.saddr;
                	p_item->dst_ip.ip = hdr->iph.daddr;
                }
                break;

            case AVM_PA_IPV6:
                dbg("doing ipv6 session");
            	p_item->flags |= SESSION_IS_IPV6;
                aass |= AASS_INGRESS_IPV6;
                if ( aass & AASS_INGRESS_IPV4 ){
                	dbg("Session seems to be 6RD, backup IPv4-addresses");
                	p_item->tun_saddr.ip = p_item->src_ip.ip;
                	p_item->tun_daddr.ip = p_item->dst_ip.ip;
                }
                memcpy(p_item->src_ip.ip6, hdr->ipv6h.saddr.s6_addr32, 16);
                memcpy(p_item->dst_ip.ip6, hdr->ipv6h.daddr.s6_addr32, 16);
                break;

            case AVM_PA_PORTS:
                p_item->src_port = ntohs( hdr->ports[0] ); 
                p_item->dst_port = ntohs( hdr->ports[1] );
                break;

            case AVM_PA_PPPOE:
                /* WAN ingress */
                p_item->pppoe_session_id = hdr->pppoeh.sid;
                ppa_debug(DBG_ENABLE_MASK_AVM_PPA, "[%s] incoming pppoe_id %d\n", __FUNCTION__, p_item->pppoe_session_id);
                p_item->flags |= SESSION_VALID_PPPOE;
                aass |= AASS_INGRESS_PPPOE;
                break;

            case AVM_PA_VLAN:
                dbg(" got incoming vlan id %d ", VLAN_ID( &hdr->vlanh ));
                if ( nr_vlan_in < num_supported_ingerss_vlan_tags ) {
                    incoming_vlan_tci[nr_vlan_in++] = htons(hdr->vlanh.vlan_tci);
                } else {
                    err("to many incoming vlan tags (>%d) ", num_supported_ingerss_vlan_tags);
                    goto ERR_FREE_PITEM;
                }
                break;

            case AVM_PA_PPP:
                dbg(" ingress PPP -> nothing to do" );
                break;

            default:
                err(" unknown ingress match type %d ", p->type);
                goto ERR_FREE_PITEM;
            	break;
        }
    }

    if (avm_session->mod.v4_mod.flags & AVM_PA_V4_MOD_DADDR) {
    	if (p_item->flags & SESSION_LAN_ENTRY){
    		err("DNAT only supported for WAN-SESSIONS flags:%x", p_item->flags);
    		goto ERR_FREE_PITEM;
    	}
        p_item->nat_ip.ip = avm_session->mod.v4_mod.daddr; 
        p_item->flags |= SESSION_VALID_NAT_IP;
    }

    if (avm_session->mod.v4_mod.flags & AVM_PA_V4_MOD_SADDR) {
    	if (p_item->flags & SESSION_WAN_ENTRY){
    		err("SNAT only supported for LAN-SESSIONS; flags:%x", p_item->flags);
    		goto ERR_FREE_PITEM;
    	}
        p_item->nat_ip.ip = avm_session->mod.v4_mod.saddr; 
        p_item->flags |= SESSION_VALID_NAT_IP;
    }

    if (avm_session->mod.v4_mod.flags & AVM_PA_V4_MOD_SPORT) {
    	if (p_item->flags & SESSION_WAN_ENTRY){
    		err("sport nat-port %d only supported for LAN-SESSIONS", p_item->nat_port);
    		goto ERR_FREE_PITEM;
    	}
    	p_item->nat_port = avm_session->mod.v4_mod.sport;
        p_item->flags |= SESSION_VALID_NAT_PORT ;
    }

    if (avm_session->mod.v4_mod.flags & AVM_PA_V4_MOD_DPORT) {
    	if (p_item->flags & SESSION_LAN_ENTRY){
    		err("dport nat-port %d only supported for WAN-SESSIONS", p_item->nat_port);
    		goto ERR_FREE_PITEM;
    	}
    	p_item->nat_port = avm_session->mod.v4_mod.dport;
        p_item->flags |= SESSION_VALID_NAT_PORT;
    }

     /*--------------------*\
      * Egress
     \*--------------------*/
    for (n=0; n < avm_session->egress[0].match.nmatch; n++) {
        struct avm_pa_match_info *p = avm_session->egress[0].match.match+n;
        hdrunion_t *hdr = (hdrunion_t *)&avm_session->egress[0].match.hdrcopy[p->offset + avm_session->egress[0].match.hdroff ];
    	int num_supported_egerss_vlan_tags = (egress_hw->flags & AVMNET_DEVICE_IFXPPA_ATM_WAN)? MAX_SUPPORTED_VLAN_TAGS : 1;


        switch (p->type) {

            case AVM_PA_VLAN:
                dbg(" egress vlan id %d ", VLAN_ID( &hdr->vlanh ));
                if ( nr_vlan_out < num_supported_egerss_vlan_tags ) {
                    outgoing_vlan_tci[nr_vlan_out++] = htons(hdr->vlanh.vlan_tci);
                } else {
                    err("to many outgoing vlan tags (>%d) ", num_supported_egerss_vlan_tags);
                    goto ERR_FREE_PITEM;
                }
                break;

            case AVM_PA_ETH:
                dbg("egress eth" );
                memcpy( &p_item->dst_mac, &hdr->ethh.h_dest, PPA_ETH_ALEN ); 
                memcpy( &p_item->new_src_mac, &hdr->ethh.h_source, PPA_ETH_ALEN );
                p_item->flags |= SESSION_VALID_NEW_SRC_MAC;
                aass |= AASS_EGRESS_ETHERNET;
                break;

            case AVM_PA_IPV4:
                dbg("egress IPV4" );
                aass |= AASS_EGRESS_IPV4;
                if ( aass & AASS_INGRESS_IPV6 ){
                	char strbuf[256];
                	p_item->tun_saddr.ip = hdr->iph.saddr;
                	p_item->tun_daddr.ip = hdr->iph.daddr;
                	dbg("egress IPV4, where ingress IPV6: tun_saddr=%s, tun_daddr=%s",
                			ppa_get_pkt_ip_string(p_item->tun_saddr, 0, strbuf),
                			ppa_get_pkt_ip_string(p_item->tun_daddr, 0, strbuf)
                	);
                }
                break;

            case AVM_PA_IPV6:
                dbg("egress IPV6" );
                aass |= AASS_EGRESS_IPV6;
                break;

            case AVM_PA_PORTS:
                dbg("egress ports -> nothing to do" );
                break;

            case AVM_PA_PPP:
            	p_item->flags |= SESSION_TX_ITF_PPPOA;
            	aass |= AASS_EGRESS_PPPOA;
                dbg("egress PPP" );
                break;

            case AVM_PA_PPPOE:
                dbg("egress PPPOE" );
                {
                	struct pppoehdr *pppoe_hdr = (struct pppoehdr *)(avm_session->egress[0].match.hdrcopy +avm_session->egress[0].match.hdroff + avm_session->egress[0].pppoe_offset);
                	p_item->pppoe_session_id = pppoe_hdr->sid;
                	ppa_debug(DBG_ENABLE_MASK_AVM_PPA, "[%s] outgoing pppoe_id %d\n", __FUNCTION__, p_item->pppoe_session_id);
                	p_item->flags |= SESSION_VALID_PPPOE;
                	aass |= AASS_EGRESS_PPPOE;
                }
                break;

            default:
                err(" unknown egress match type %d ", p->type);
                goto ERR_FREE_PITEM;
            	break;
        }
    }

     /*--------------------*\
      * Constraints checking
     \*--------------------*/

    if ( !( aass & ( AASS_EGRESS_ETHERNET | AASS_INGRESS_ETHERNET )) ){
        err("no valid ethernet - we need at least one device to be ethernet");
        goto ERR_FREE_PITEM;
    }

    if ( ( aass & AASS_INGRESS_IPV4 ) && (aass & AASS_EGRESS_IPV6) )  {
    	if (aass & AASS_INGRESS_IPV6){
    		dbg("6rd is supported");
    		p_item->flags |= SESSION_TUNNEL_6RD;
    	} else {
    		err("ipv4 -> ipv6 not supported");
    		goto ERR_FREE_PITEM;
    	}
    }

    if ( ( aass & AASS_INGRESS_IPV6 ) && (aass & AASS_EGRESS_IPV4) ) {
    	if (aass & AASS_EGRESS_IPV6){
    		dbg("6rd is supported");
    		p_item->flags |= SESSION_TUNNEL_6RD;
    	} else {
    		err("ipv6 -> ipv4 not supported");
    		goto ERR_FREE_PITEM;
    	}
    }

    if ( ( p_item->flags & SESSION_LAN_ENTRY )  && (!( aass & (AASS_EGRESS_ETHERNET | AASS_EGRESS_PPPOA | AASS_EGRESS_PPPOE) )) ) {
    	dbg("egress fallback: session has to be IPoA" );
        p_item->flags |= SESSION_TX_ITF_IPOA ;
        aass |= AASS_EGRESS_IPOA;
    }
    /*--------------------*\
     *     setup vlan
    \*--------------------*/

    if ( nr_vlan_out || nr_vlan_in ) {
        ppa_debug(DBG_ENABLE_MASK_AVM_PPA, "[%s] NR VLAN (in=%d, out=%d)\n", __FUNCTION__, nr_vlan_in , nr_vlan_out);

        if ( nr_vlan_out > 0 ) {
            p_item->out_vlan_tag = ( (ETH_P_8021Q << 16) | outgoing_vlan_tci[0]);
            p_item->flags |= SESSION_VALID_OUT_VLAN_INS;
        }

       	if ( nr_vlan_out > 1 ) {
            p_item->new_vci = outgoing_vlan_tci[1];
            p_item->flags |= SESSION_VALID_VLAN_INS;
       	}

       	if ( nr_vlan_in == 2 ){
       	    // In ATM this can not happen, because of num_supported_egerss_vlan_tags
            p_item->flags |= SESSION_VALID_VLAN_RM;
            p_item->flags |= SESSION_VALID_OUT_VLAN_RM;
       	}

       	if ( nr_vlan_in == 1 ) {
       		if (ingress_hw->flags & AVMNET_DEVICE_IFXPPA_ATM_WAN)
       			p_item->flags |= SESSION_VALID_VLAN_RM;
       		else
       			p_item->flags |= SESSION_VALID_OUT_VLAN_RM;
       	}
        ppa_debug(DBG_ENABLE_MASK_AVM_PPA, "[%s] flags=%#x (vlan_tag=%#x, new_vci=%#x)\n", __FUNCTION__, p_item->flags, p_item->out_vlan_tag, ((unsigned int)p_item->new_vci & 0xFFFF));
    }

    dump_list_item(p_item, "avm_pa_add_ifx_session (after setting)");

    if ( !(aass & AASS_UPDATE_OLD_SESSION )){
        ppa_insert_session_item(p_item);
        aass |= AASS_UPDATE_OLD_SESSION;
    }

    if ( (hw_ret = ppa_hw_add_session(p_item, 0)) != IFX_SUCCESS ) {
        ppa_debug(DBG_ENABLE_MASK_AVM_PPA, "[%s] ppa_hw_add_session failed: %d (probably no more space for any sessions)\n", __FUNCTION__, hw_ret);
        goto ERR_FREE_PITEM;
    }

    avm_session->hw_session = p_item;

   	ppa_debug(DBG_ENABLE_MASK_VDEV_SESSION, "%lu: add routing_entry=%d (%s)%s\n", jiffies,
   	        p_item->routing_entry & 0x7FFFFFFF,
   	        ( p_item->flags & SESSION_LAN_ENTRY )?"LAN":"WAN",
   	        ( p_item->flags & SESSION_IS_IPV6)?"IPV6":"");
    return AVM_PA_TX_SESSION_ADDED; //AVM_PA_SUCCESS;

ERR_FREE_PITEM:
    /*--- cannot use ppa_free_session_list_item, as session is not added in hw ---*/
    if ( aass & AASS_UPDATE_OLD_SESSION )
        ppa_remove_session_item(p_item);
    ppa_mem_cache_free(p_item, g_session_item_cache);

ERR:

	// print_flags( aass );
    return AVM_PA_TX_ERROR_SESSION; //AVM_PA_FAILURE;
}

/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/

int avm_pa_remove_ifx_session ( struct avm_pa_session *avm_session ) {
    
	struct session_list_item *p_item = avm_session->hw_session;
	unsigned int store_routing_entry = p_item->routing_entry;
	unsigned int store_is_lan = p_item->flags & SESSION_LAN_ENTRY;

    dump_list_item(p_item, "ifx_ppa_session_delete");
    
    ppa_debug( DBG_ENABLE_MASK_AVM_PPA, "[%s] remove session %p\n", __FUNCTION__, p_item );

    ppa_remove_session_item(p_item);

    ppa_hw_del_session(p_item);

    ifx_ppa_pwm_deactivate_module();

   	ppa_debug(DBG_ENABLE_MASK_VDEV_SESSION, "%lu: remove routing_entry=%d (%s)\n", jiffies, store_routing_entry & 0x7FFFFFFF,
   	        store_is_lan?"LAN":"WAN");
    avm_sessionh_table_remove( store_routing_entry, store_is_lan );

    ppa_mem_cache_free(p_item, g_session_item_cache);

    ppa_debug( DBG_ENABLE_MASK_AVM_PPA, "[%s] deaktivate ppa pwm\n", __FUNCTION__ );

	return IFX_SUCCESS;
}

/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/

uint64_t avm_pa_get_ifx_bytes(unsigned int session_id){
	
    struct session_list_item *p_item = NULL;
    ppa_debug(DBG_ENABLE_MASK_AVM_PPA, "[%s] id=%d\n", __FUNCTION__, session_id);

    if ( avm_lookup_session(session_id, &p_item) != IFX_PPA_SESSION_EXISTS ){
    	ppa_debug(DBG_ENABLE_MASK_AVM_PPA, "[%s]session not found!\n", __FUNCTION__);
        return 0;
    }
    ppa_debug(DBG_ENABLE_MASK_AVM_PPA, "[%s]\n", __FUNCTION__);
    return p_item->acc_bytes;
}

/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/

static struct avm_hardware_pa_instance ifx_pa = {
    .add_session = avm_pa_add_ifx_session,
    .remove_session = avm_pa_remove_ifx_session,
    .name = "ifx_ppa",
};

/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/

void avm_pa_register_ifx_hw_ppa(void){
    avm_pa_multiplexer_register_instance( &ifx_pa );
    ifx_lookup_routing_entry = avm_session_handle_fast_lookup;
}

/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/

void avm_pa_unregister_ifx_hw_ppa(void){
    avm_pa_multiplexer_unregister_instance( &ifx_pa );
    ifx_lookup_routing_entry = NULL;
}

/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/

EXPORT_SYMBOL(avm_pa_register_ifx_hw_ppa);
EXPORT_SYMBOL(avm_pa_unregister_ifx_hw_ppa);


#endif
    

EXPORT_SYMBOL(ppa_bridging_session_start_iteration);

#if defined(CONFIG_AVM_IFX_PA_ADD_SESSION_PROC)
EXPORT_SYMBOL(proc_write_avm_add_session);
EXPORT_SYMBOL(proc_write_avm_del_session);
#endif
EXPORT_SYMBOL(ppa_bridging_session_iterate_next);
EXPORT_SYMBOL(ppa_bridging_session_stop_iteration);
EXPORT_SYMBOL(ppa_session_start_iteration);
EXPORT_SYMBOL(ppa_session_iterate_next);
EXPORT_SYMBOL(ppa_session_stop_iteration);
EXPORT_SYMBOL(ppa_mc_group_start_iteration);
EXPORT_SYMBOL(ppa_mc_group_iterate_next);
EXPORT_SYMBOL(ppa_mc_group_stop_iteration);
EXPORT_SYMBOL(ppa_alloc_session_list_item);
EXPORT_SYMBOL(ppa_free_session_list_item);
EXPORT_SYMBOL(ppa_insert_session_item);
EXPORT_SYMBOL(ppa_api_get_session_poll_timer);
EXPORT_SYMBOL(global_session_count);
EXPORT_SYMBOL(global_fail_count);


